"""
EEmiLib (Electron EMIssion Library) holds several electron emission models and
offers a simple way to fit the on electron emission data.

"""

import importlib.metadata
from importlib import resources

from eemilib.util.log_manager import set_up_logging

DOC_URL = "https://eemilib.readthedocs.io/en/latest"
__version__ = importlib.metadata.version("eemilib")

teey_cu = resources.files("eemilib.data.cu.emission_yield")
emission_energy_ag = resources.files("eemilib.data.ag.emission_energy")

set_up_logging("EEmiLib")



"""Define a model parameter."""

import numpy as np


class Parameter:
    """An electron emission model parameter."""

    _tol: float = 1e-10

    def __init__(
        self,
        markdown: str,
        unit: str = "1",
        value: float = 0.0,
        *,
        lower_bound: float = -np.inf,
        upper_bound: float = np.inf,
        description: str = "",
        is_locked: bool = False,
    ) -> None:
        """Instantiate the parameter.

        Parameters
        ----------
        markdown :
            The name of the parameter, in markdown format.
        unit :
            The unit of the parameter. The default is ``"1"`` (no unit).
        value :
            A first value for the parameter. The default is ``0.``.
        lower_bound :
            A first lower bound for the parameter. The default is ``-np.inf``.
        upper_bound :
            A first upper bound for the parameter. The default is ``np.inf``.
        description :
            A description string for the parameter. The default is an empty
            string.
        is_locked :
            Forces the parameters to a certain value, which will not be
            modified by EEmiLib.

        """
        self.markdown = markdown
        self.unit = unit
        self._value = value
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self.description = description
        self.is_locked = is_locked

    def __repr__(self) -> str:
        """Print out name of parameter and current value."""
        return f"{self.name} ({self.unit}): {self.value} {self.description}"

    def __str__(self) -> str:
        """Return name of parameter, its value and its unit."""
        return f"{self.value:.3f} [{self.unit}]"

    @property
    def name(self) -> str:
        """Return markdown name of parameter with its unit."""
        return f"{self.markdown} [{self.unit}]"

    @property
    def value(self) -> float:
        """Give the current value of the parameter."""
        return self._value

    @value.setter
    def value(self, value: float) -> None:
        """Set the value of the parameter."""
        self._value = value
        if not self.is_locked:
            return
        self._set_tiny_bounds()

    @property
    def lower_bound(self) -> float:
        """Give the current lower bound of the parameter."""
        return self._lower_bound

    @lower_bound.setter
    def lower_bound(self, lower_bound: float) -> None:
        """Set the lower bound of the parameter."""
        self._lower_bound = lower_bound
        return

    @property
    def upper_bound(self) -> float:
        """Give the current upper bound of the parameter."""
        return self._upper_bound

    @upper_bound.setter
    def upper_bound(self, upper_bound: float) -> None:
        """Set the upper bound of the parameter."""
        self._upper_bound = upper_bound
        return

    def lock(self) -> None:
        """Set the parameter to its current value."""
        self.is_locked = True
        self._set_tiny_bounds()

    def unlock(self) -> None:
        """Allow parameter to be changed again during optimisation."""
        self.is_locked = False
        self.lower_bound = -np.inf
        self.upper_bound = +np.inf

    def _set_tiny_bounds(self) -> None:
        """Set tiny bounds around :attr:`value`."""
        bound_1, bound_2 = self.value - self._tol, self.value + self._tol
        self.lower_bound = min(bound_1, bound_2)
        self.upper_bound = max(bound_1, bound_2)



"""Define a structure to select a :class:`.Model` mandatory files."""

from collections.abc import Collection
from dataclasses import dataclass

from eemilib.util.constants import (
    IMPLEMENTED_EMISSION_DATA,
    ImplementedEmissionData,
    ImplementedPop,
)


@dataclass
class ModelConfig:
    """Define mandatory files for a :class:`.Model`."""

    emission_yield_files: Collection[ImplementedPop]
    emission_energy_files: Collection[ImplementedPop]
    emission_angle_files: Collection[ImplementedPop]

    def mandatory_populations(
        self, emission_data_type: ImplementedEmissionData
    ) -> list[ImplementedPop]:
        """Tell which population should be given in files, for a data type."""
        if emission_data_type == "Emission Yield":
            return list(self.emission_yield_files)
        if emission_data_type == "Emission Energy":
            return list(self.emission_energy_files)
        if emission_data_type == "Emission Angle":
            return list(self.emission_angle_files)
        raise RuntimeError(
            f"{emission_data_type = } is not in {IMPLEMENTED_EMISSION_DATA = }"
        )



"""
Emission energy measurements realized at ONERA, Toulouse, France using the
DEESSE test facility on 2018-06-04.

The Ag sample was heated at 200 celsius during 2 hours.

It is called "heated Ag#1" in my PhD :cite:`Placais2021`. Emission energy data
was corrected using M. Villemant procedure :cite:`Villemant2018`.

This data can be loaded with :class:`.pandas_loader.PandasLoader`.

"""



"""Define an object to store an emission yield."""

import logging
from pathlib import Path
from typing import Self

import pandas as pd
from eemilib.emission_data.emission_data import EmissionData
from eemilib.emission_data.helper import (
    get_crossover_energies,
    get_emax_eymax,
    resample,
)
from eemilib.loader.loader import Loader
from eemilib.plotter.plotter import Plotter
from eemilib.util.constants import (
    ImplementedPop,
    col_energy,
    col_normal,
    md_ey,
)


class EmissionYield(EmissionData):
    """An emission yield."""

    def __init__(self, population: ImplementedPop, data: pd.DataFrame) -> None:
        """Instantiate the data.

        Parameters
        ----------
        population :
            The concerned population of electrons.
        data :
            Structure holding the data. Must have an ``Energy (eV)`` column
            holding PEs energy. And one or several columns ``theta [deg]``,
            where ``theta`` is the value of the incidence angle and content is
            corresponding emission yield.

        """
        super().__init__(population, data)
        self.energies = data[col_energy].to_numpy()
        self.angles = [
            float(col.split()[0]) for col in data.columns if col != col_energy
        ]
        #: Energy at the maximum emission yield in :unit:`eV`. Not defined for
        #: BEs.
        self.e_max: float
        #: Maximum emission yield. Not defined for BEs.
        self.ey_max: float
        #: First cross-over enrergy in :unit:`eV`. Not defined for BEs.
        self.e_c1: float
        #: Second cross-over enrergy in :unit:`eV`. Not defined for BEs.
        self.e_c2: float | None
        if self.population in ("SE", "all"):
            self.e_max, self.ey_max, self.e_c1, self.e_c2 = self._parameters(
                n_resample=1000
            )

    @classmethod
    def from_filepath(
        cls,
        population: ImplementedPop,
        loader: Loader,
        *filepath: str | Path,
    ) -> Self:
        """Instantiate the data from files.

        Parameters
        ----------
        loader :
            The object that will load the data.
        population :
            The concerned population of electrons.
        *filepath :
            Path(s) to file holding data under study.

        """
        data = loader.load_emission_yield(*filepath)
        return cls(population, data)

    @property
    def label(self) -> str:
        """Print nature of data (markdown)."""
        return md_ey[self.population]

    def _parameters(
        self,
        n_resample: int = -1,
    ) -> tuple[float, float, float, float | None]:
        """Compute the characteristics of the emission yield."""
        assert 0.0 in self.angles, "Need the normal incidence measurements."

        normal_ey = self.data[[col_energy, col_normal]]
        assert isinstance(normal_ey, pd.DataFrame)
        normal_ey = resample(normal_ey, n_resample)

        e_max, sigma_max = self._get_maximum_ey(normal_ey)
        e_c1, e_c2 = self._get_crossovers(normal_ey, e_max)
        return e_max, sigma_max, e_c1, e_c2

    def _get_maximum_ey(
        self, normal_ey: pd.DataFrame, tol_energy: float = 10.0
    ) -> tuple[float, float]:
        r"""Get the position and value of max emission yield.

        Parameters
        ----------
        normal_ey :
            Holds energy of PEs as well as emission yield at nominal incidence.
        tol_energy :
            If the :math:`E_{max}` is too close to the maximum PE energy, an
            warning is raised; tolerance is ``tol_energy``. The default value
            is 10 eV.

        Returns
        -------
            :math:`E_{max}` and :math:`\sigma_{max}`.
        """
        e_max, sigma_max = get_emax_eymax(normal_ey)
        if abs(e_max - self.energies[-1]) < tol_energy:
            logging.warning(
                "E_max is very close to the last measured energy. Maybe "
                "maximum emission yield was not reached?"
            )
        return e_max, sigma_max

    def _get_crossovers(
        self,
        normal_ey: pd.DataFrame,
        e_max: float,
        min_e: float = 10.0,
        tol_ey: float = 0.01,
    ) -> tuple[float, float | None]:
        """Compute first and second crossover energies.

        Parameters
        ----------
        normal_ey :
            Holds energy of PEs as well as emission yield at nominal incidence.
        e_max :
            Energy of maximum emission yield. Used to discriminate
            :math:`E_{c1}` from :math:`E_{c2}`.
        min_e :
            Energy under which :math:`E_{c1}` is not searched. It is useful if
            emission yield data comes from a model which sets the emission
            yield to unity at very low energies (eg some implementations of
            Vaughan). The default value is 10 eV.
        tol_ey :
            It the emission yield is too far from unity at crossover energy, a
            warning is raised. Tolerance is ``tol_ey``. The default value is
            ``0.01``.

        Returns
        -------
        tuple[float, float | None]
            First and second crossover energies.

        """
        (ec1, ey_ec1), (ec2, ey_ec2) = get_crossover_energies(
            normal_ey, e_max, min_e
        )
        if abs(ey_ec1 - 1.0) > tol_ey:
            logging.warning(
                f"The emission yield at first crossover energy is {ey_ec1}, "
                "which is far from unity. Keeping it anyway."
            )

        if abs(ey_ec2 - 1.0) > tol_ey:
            logging.info(
                f"The emission yield at second crossover energy is {ey_ec2}, "
                "which is far from unity. Maybe its energy lies outside of the"
                " measurement range. Setting E_c2 = None."
            )
            ec2 = None

        return ec1, ec2

    def plot[T](
        self,
        plotter: Plotter,
        *args,
        lw: float | None = 0.0,
        marker: str | None = "+",
        axes: T | None = None,
        grid: bool = True,
        population: ImplementedPop | None = None,
        **kwargs,
    ) -> T:
        """Plot the contained data using plotter.

        This wrapper simply calls the :meth:`.Plotter.plot_emission_yield`
        method.

        """
        return plotter.plot_emission_yield(
            df=self.data,
            *args,
            axes=axes,
            lw=lw,
            marker=marker,
            grid=grid,
            label=self.label,
            population=population,
            **kwargs,
        )



#!/usr/bin/env python3
"""Define a GUI.

.. todo::
    Export/Import settings

.. todo::
    logging module

.. todo::
    Add measurables at bottom

"""
import importlib
import logging
import sys
from abc import ABCMeta
from types import ModuleType
from typing import Literal

import numpy as np
from eemilib.core.model_config import ModelConfig
from eemilib.emission_data.data_matrix import DataMatrix
from eemilib.gui.file_selection import file_selection_matrix
from eemilib.gui.helper import (
    PARAMETER_ATTR_TO_POS,
    PARAMETER_POS_TO_ATTR,
    format_number,
    set_dropdown_value,
    set_help_button_action,
    setup_dropdown,
    setup_linspace_entries,
    setup_lock_checkbox,
    to_plot_checkboxes,
)
from eemilib.gui.model_selection import model_configuration
from eemilib.loader.loader import Loader
from eemilib.model.model import Model
from eemilib.plotter.plotter import Plotter
from eemilib.util.constants import (
    IMPLEMENTED_EMISSION_DATA,
    IMPLEMENTED_POP,
    ImplementedEmissionData,
    ImplementedPop,
)
from PyQt5.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QGroupBox,
    QLineEdit,
    QListWidget,
    QMainWindow,
    QPushButton,
    QRadioButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

DROPDOWNS = ("Loader", "Model", "Plotter")
Dropdowns = Literal["Loader", "Model", "Plotter"]


class MainWindow(QMainWindow):
    """This object holds the GUI."""

    #: If selecting Model in dropdown should automatically fill the appopriate
    #: data to plot checkbox
    autofill_data_to_plot = True
    #: If selecting Model in dropdown should automatically fill the appopriate
    #: emission data checkbox
    autofill_nature_to_plot = True
    #: If loading data should automatically fill the energy/angle ranges with
    #: their maximum values
    autofill_plotting_ranges = True

    def __init__(
        self,
        default_model: str = "Vaughan",
        default_loader: str = "PandasLoader",
        default_plotter: str = "PandasPlotter",
    ) -> None:
        """Create the GUI."""
        self._defaults: dict[Dropdowns, str] = {
            "Model": default_model,
            "Loader": default_loader,
            "Plotter": default_plotter,
        }
        # EEmiLib attributes
        self.data_matrix = DataMatrix()
        self.loader: Loader
        self.model: Model
        self.axes = None

        super().__init__()
        self.setWindowTitle("EEmiLib")

        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)

        self.main_layout = QVBoxLayout(self.central_widget)

        self.file_lists = self.setup_file_selection_matrix()

        self.dropdowns: dict[str, QComboBox] = {}

        self.loader_classes: dict[str, str]
        self.loader_help_button: QPushButton
        self.setup_loader_dropdown()

        self.model_classes: dict[str, str]
        self.model_help_button: QPushButton
        self.setup_model_dropdown()

        self.model_table = self.setup_model_configuration()
        self.energy_angle_group: QGroupBox
        self.energy_angle_layout: QVBoxLayout
        self.last_energy_widget: QLineEdit
        self.last_theta_widget: QLineEdit
        self.n_theta_widget: QLineEdit
        self.setup_energy_angle_inputs()

        self.plotter_classes: dict[str, str]
        self.plot_measured_button: QPushButton
        self.plot_model_button: QPushButton
        self.data_checkboxes: list[QRadioButton]
        self.population_checkboxes: list[QCheckBox]
        self.setup_plotter_dropdowns()

        # Call the methods called by the model_dropdown index change
        self._set_default_dropdown()

    # =========================================================================
    # File selection
    # =========================================================================
    def setup_file_selection_matrix(self) -> list[list[None | QListWidget]]:
        """Create the 4 * 3 matrix to select the files to load."""
        file_matrix_group, file_lists = file_selection_matrix(self)
        self.main_layout.addWidget(file_matrix_group)
        return file_lists

    def _deactivate_unnecessary_file_widgets(self) -> None:
        """Grey out the files not needed by current model."""
        model = self._dropdown_to_class("Model")()
        if not isinstance(model, Model):
            return
        config: ModelConfig = model.model_config

        # Get required file types for each population type
        required_files = {
            "Emission Yield": config.emission_yield_files,
            "Emission Energy": config.emission_energy_files,
            "Emission Angle": config.emission_angle_files,
        }

        for i, pop in enumerate(IMPLEMENTED_POP):
            for j, data_type in enumerate(IMPLEMENTED_EMISSION_DATA):
                is_required = pop in required_files.get(data_type, [])
                self._set_list_widget_state(self.file_lists[i][j], is_required)

    # =========================================================================
    # Load files
    # =========================================================================
    def setup_loader_dropdown(self) -> None:
        """Set the :class:`.Loader` related interface."""
        classes, layout, dropdown, buttons = setup_dropdown(
            module_name="eemilib.loader",
            base_class=Loader,
            buttons_args={
                "Help": lambda _: logging.info("Help not set."),
                "Load data": self.load_data,
            },
        )
        self.loader_classes = classes
        dropdown.currentIndexChanged.connect(self._setup_loader)
        dropdown.setCurrentText
        self.dropdowns["Loader"] = dropdown
        self.loader_help_button = buttons[0]
        self.main_layout.addLayout(layout)
        return

    def _setup_loader(self) -> None:
        """Setup new loader whenever the dropdown menu is changed."""
        self.loader = self._dropdown_to_class("Loader")()
        set_help_button_action(self.loader_help_button, self.loader)

    def load_data(self) -> None:
        """Load all the files set in GUI."""
        for i in range(len(IMPLEMENTED_POP)):
            for j in range(len(IMPLEMENTED_EMISSION_DATA)):
                file_list_widget = self.file_lists[i][j]
                if file_list_widget is not None:
                    file_names = [
                        file_list_widget.item(k).text()
                        for k in range(file_list_widget.count())
                    ]
                    self.data_matrix.set_files(file_names, row=i, col=j)

        try:
            self.data_matrix.load_data(self.loader)
        except Exception as e:
            logging.error(
                "An error was raised during the loading of the data file. "
                "Check that the format of the files is consistent with what "
                f"is expected by the data loader. Error message:\n{e}"
            )

        if self.autofill_plotting_ranges:
            self._fill_plotting_ranges()

    # =========================================================================
    # Model
    # =========================================================================
    def setup_model_dropdown(self) -> None:
        """Set the :class:`.Model` related interface.

        Assign the ``model_classes`` and ``model_dropdown``.

        """
        classes, layout, dropdown, buttons = setup_dropdown(
            module_name="eemilib.model",
            base_class=Model,
            buttons_args={
                "Help": lambda _: logging.info("Help not set"),
                "Fit!": self.fit_model,
            },
        )
        self.model_classes = classes
        self.dropdowns["Model"] = dropdown
        dropdown.currentIndexChanged.connect(self._setup_model)
        dropdown.currentIndexChanged.connect(
            self._deactivate_unnecessary_file_widgets
        )
        dropdown.currentIndexChanged.connect(
            self._fill_plot_nature_and_population
        )

        self.model_help_button = buttons[0]
        self.main_layout.addLayout(layout)
        return

    def setup_model_configuration(self) -> QTableWidget:
        """Set the interface related to the model specific parameters."""
        group, model_table = model_configuration()
        self.main_layout.addWidget(group)

        return model_table

    def _setup_model(self) -> None:
        """Instantiate :class:`.Model` when it is selected in dropdown menu."""
        self.model = self._dropdown_to_class("Model")()

        set_help_button_action(self.model_help_button, self.model)

        self._populate_parameters_table_constants()
        self.model_table.itemChanged.connect(
            self._update_parameter_value_from_table
        )

    def _populate_parameters_table_constants(self) -> None:
        """Print out the model parameters in dedicated table."""
        self.model_table.setRowCount(0)

        for row, (name, param) in enumerate(self.model.parameters.items()):
            self.model_table.insertRow(row)

            self.model_table.setItem(row, 0, QTableWidgetItem(name))
            for attr in ("unit", "lower_bound", "upper_bound"):
                col = PARAMETER_ATTR_TO_POS[attr]
                attr_value = getattr(param, attr, None)
                self.model_table.setItem(
                    row, col, QTableWidgetItem(str(attr_value))
                )
            col_lock = PARAMETER_ATTR_TO_POS["lock"]
            checkbox_widget = setup_lock_checkbox(param)
            self.model_table.setCellWidget(row, col_lock, checkbox_widget)

    def _update_parameter_value_from_table(
        self, item: QTableWidgetItem
    ) -> None:
        """Update :class:`.Parameter` value based on user input in table."""
        row, col = item.row(), item.column()
        updatable_attr = ("value", "lower_bound", "upper_bound")
        attr = PARAMETER_POS_TO_ATTR[col]
        if attr not in updatable_attr:
            return

        name = self.model_table.item(row, 0).text()
        parameter = self.model.parameters.get(name)

        if parameter:
            try:
                new_value = float(item.text())
                setattr(parameter, attr, new_value)

            except ValueError:
                logging.warning(f"Invalid value entered for {name}")
                item.setText(str(parameter.value))

    def fit_model(self) -> None:
        """Perform the fit on the loaded data."""
        if not hasattr(self, "model") or not self.model:
            logging.info("Please select a model before fitting.")
            return
        self.model.find_optimal_parameters(self.data_matrix)
        self._populate_parameters_table_values()

    def _populate_parameters_table_values(self) -> None:
        """Print out the values of the model parameters in dedicated table."""
        for row, param in enumerate(self.model.parameters.values()):
            for attr in ("value",):
                col = PARAMETER_ATTR_TO_POS[attr]
                attr_value = getattr(param, attr, None)
                self.model_table.setItem(
                    row, col, QTableWidgetItem(str(attr_value))
                )

        for i, param in enumerate(self.model.parameters.values()):
            self.model_table.setItem(
                i, 2, QTableWidgetItem(format_number(param.value))
            )

    def _fill_plot_nature_and_population(self) -> None:
        """Check emission data type and population.

        When model is updated, check the ``Data to plot`` and ``Population to
        plot`` checkboxes in the ``Plot`` section that are concerned by current
        model.

        """
        try:
            model = self.model
        except AttributeError as e:
            logging.debug(
                "Model is not set, cannot fill plot nature or population "
                f"checkboxes.\n{e}"
            )
            return

        data_type_to_plot = model.emission_data_types[0]
        if self.autofill_data_to_plot:
            index = IMPLEMENTED_EMISSION_DATA.index(data_type_to_plot)
            self.data_checkboxes[index].setChecked(True)

        if self.autofill_nature_to_plot:
            pop_to_plot = set(
                model.model_config.mandatory_populations(
                    emission_data_type=data_type_to_plot
                )
                + list(model.populations)
            )
            for button, population in zip(
                self.population_checkboxes, IMPLEMENTED_POP, strict=True
            ):
                if population in pop_to_plot:
                    button.setChecked(True)
                    continue
                button.setChecked(False)

    # =========================================================================
    # Plot
    # =========================================================================
    def setup_energy_angle_inputs(self) -> None:
        """Set the energy and angle inputs for the model plot."""
        self.energy_angle_group = QGroupBox(
            "Energy and angle range (used for model plot)"
        )
        self.energy_angle_layout = QVBoxLayout()

        quantities = ("energy", "angle")
        labels = ("Energy [eV]", "Angle [deg]")
        initial_values = ((0.0, 500.0, 501), (0.0, 60.0, 4))
        max_values = (None, 90.0)
        for qty, label, initial, max_val in zip(
            quantities, labels, initial_values, max_values
        ):
            layout, first, last, points = setup_linspace_entries(
                label,
                initial_values=initial,
                max_value=max_val,
            )
            self.energy_angle_layout.addLayout(layout)
            if qty == ("energy"):
                self.last_energy_widget = last
            elif qty == ("angle"):
                self.last_theta_widget = last
                self.n_theta_widget = points

            for attr, attr_name in zip(
                (first, last, points), ("first", "last", "points")
            ):
                setattr(self, "_".join((qty, attr_name)), attr)

        self.energy_angle_group.setLayout(self.energy_angle_layout)
        self.main_layout.addWidget(self.energy_angle_group)

    def setup_plotter_dropdowns(self) -> None:
        """Set the :class:`.Plotter` related interface."""
        self._set_up_data_to_plot_checkboxes()
        self._set_up_population_to_plot_checkboxes()

        classes, layout, dropdown, buttons = setup_dropdown(
            module_name="eemilib.plotter",
            base_class=Plotter,
            buttons_args={
                "Plot file": self.plot_measured,
                "Plot model": self.plot_model,
                "New figure": lambda _: setattr(self, "axes", None),
            },
        )
        self.plotter_classes = classes
        self.main_layout.addLayout(layout)
        self.dropdowns["Plotter"] = dropdown
        self.plot_measured_button = buttons[0]
        self.plot_model_button = buttons[1]

    def _set_up_data_to_plot_checkboxes(self) -> None:
        """Add checkbox to select which data should be plotted."""
        layout, checkboxes = to_plot_checkboxes(
            "Data to plot:",
            IMPLEMENTED_EMISSION_DATA,
            several_can_be_checked=False,
        )
        self.main_layout.addLayout(layout)
        self.data_checkboxes = checkboxes

    def _set_up_population_to_plot_checkboxes(self) -> None:
        """Add checkbox to select which population should be plotted."""
        layout, checkboxes = to_plot_checkboxes(
            "Population to plot:",
            IMPLEMENTED_POP,
            several_can_be_checked=True,
        )
        self.main_layout.addLayout(layout)
        self.population_checkboxes = checkboxes

    def plot_measured(self) -> None:
        """Plot the desired data, as imported."""
        plotter = self._dropdown_to_class("Plotter")(gui=True)

        success_pop, populations = self._get_populations_to_plot()
        if not success_pop:
            return
        success_data, emission_data_type = (
            self._get_emission_data_type_to_plot()
        )
        if not success_data:
            return

        self.axes = self.data_matrix.plot(
            plotter,
            population=populations,
            emission_data_type=emission_data_type,
            axes=self.axes,
        )

    def plot_model(self) -> None:
        """Plot the desired data, as modelled."""
        plotter = self._dropdown_to_class("Plotter")(gui=True)

        success_pop, populations = self._get_populations_to_plot()
        if not success_pop:
            return
        success_data, emission_data_type = (
            self._get_emission_data_type_to_plot()
        )
        if not success_data:
            return
        success_ene, energies = self._gen_linspace("energy")
        if not success_ene:
            return
        success_angle, angles = self._gen_linspace("angle")
        if not success_angle:
            return

        self.axes = self.model.plot(
            plotter,
            population=populations,
            emission_data_type=emission_data_type,
            energies=energies,
            angles=angles,
            axes=self.axes,
        )

    def _get_emission_data_type_to_plot(
        self,
    ) -> tuple[bool, ImplementedEmissionData | None]:
        """Read input to determine the emission data type to plot."""
        emission_data_type = [
            IMPLEMENTED_EMISSION_DATA[i]
            for i, checked in enumerate(self.data_checkboxes)
            if checked.isChecked()
        ]
        if len(emission_data_type) == 0:
            logging.error("Please provide a type of data to plot.")
            return False, None
        return True, emission_data_type[0]

    def _get_populations_to_plot(self) -> tuple[bool, list[ImplementedPop]]:
        """Read input to determine the populations to plot."""
        success = True
        populations = [
            IMPLEMENTED_POP[i]
            for i, checked in enumerate(self.population_checkboxes)
            if checked.isChecked()
        ]
        if len(populations) == 0:
            logging.error("Please provide at least one population to plot.")
            success = False
        return success, populations

    def _gen_linspace(
        self, variable: Literal["energy", "angle"]
    ) -> tuple[bool, np.ndarray]:
        """Take the desired input, check validity, create array of values."""
        success = True
        linspace_args = []
        for box in ("first", "last", "points"):
            line_name = "_".join((variable, box))
            qline_edit = getattr(self, line_name, None)
            if qline_edit is None:
                logging.error(f"The attribute {line_name} is not defined.")
                success = False
                continue

            assert isinstance(qline_edit, QLineEdit)
            value = qline_edit.displayText()
            if not value:
                logging.error(f"You must give a value in {line_name}.")
                success = False
                continue
            linspace_args.append(value)

        if not success:
            return success, np.linspace(0, 10, 11)

        return success, np.linspace(
            float(linspace_args[0]),
            float(linspace_args[1]),
            int(linspace_args[2]),
        )

    def _fill_plotting_ranges(self) -> None:
        """Fill energy and angle plotting ranges to match data files values.

        This method is called when the button ``Load`` is pressed.

        """
        try:
            model = self.model
        except AttributeError as e:
            logging.debug(
                "Model is not set, cannot fill energy/angle plotting ranges. "
                f"\n{e}"
            )
            return
        try:
            data_matrix = self.data_matrix
        except AttributeError as e:
            logging.debug(
                "DataMatrix is not set, cannot fill energy/angle plotting "
                f"ranges.\n{e}"
            )
            return

        if not self.autofill_plotting_ranges:
            return
        data_type_to_plot = model.emission_data_types[0]

        data = data_matrix.get_data(emission_data_type=data_type_to_plot)
        if len(data) == 0:
            logging.debug(
                "No valid data, cannot fill energy/angle plotting ranges."
            )
            return
        data_subset = data[0]

        e_maxi = max(data_subset.energies)
        if e_maxi is not None and not np.isnan(e_maxi):
            logging.debug(f"Setting {e_maxi = }")
            self.last_energy_widget.setText(str(e_maxi))

        theta_maxi = max(data_subset.angles)
        n_theta = len(data_subset.angles)
        if theta_maxi is not None and not np.isnan(theta_maxi):
            logging.debug(f"Setting {theta_maxi = }")
            self.last_theta_widget.setText(str(theta_maxi))
            logging.debug(f"Setting {n_theta = }")
            self.n_theta_widget.setText(str(n_theta))

    # =========================================================================
    # Helper
    # =========================================================================
    def _dropdown_to_class(self, name: Dropdowns) -> ABCMeta:
        """Convert dropdown entry to class."""
        dropdown = self.dropdowns.get(name, None)
        assert dropdown is not None, f" The dropdown {name} is not defined."

        module_names_to_paths = "_".join((name.lower(), "classes"))
        module_name_to_path = getattr(self, module_names_to_paths, None)
        assert module_name_to_path is not None, (
            f"The dictionary {module_names_to_paths}, linking every module"
            " name to its path, is not defined."
        )

        selected: str = dropdown.currentText()
        module_path: str = module_name_to_path[selected]
        module: ModuleType = importlib.import_module(module_path)
        my_class = getattr(module, selected)
        return my_class

    def _set_list_widget_state(
        self, widget: QListWidget, enabled: bool
    ) -> None:
        """Enable or disable a QListWidget based on ``enabled``."""
        if enabled:
            widget.setStyleSheet("background-color: white;")
            widget.setEnabled(True)
            return
        widget.setStyleSheet("background-color: lightgray;")
        widget.setEnabled(False)

    # =========================================================================
    # Misc
    # =========================================================================
    def _set_default_dropdown(self) -> None:
        """Set dropdowns to their default values.

        We call this method at the end of the GUI initialization rather than
        at the creation of the dropdowns to ensure that every side effects
        is executed.

        """
        for key in DROPDOWNS:
            set_dropdown_value(self.dropdowns[key], self._defaults[key])


def main():
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



"""Define generic utility functions."""

import inspect
import pkgutil
from abc import ABCMeta
from collections.abc import Iterable, Iterator
from typing import Any

from eemilib import DOC_URL


def get_classes(module_name: str, base_class: ABCMeta) -> dict[str, str]:
    """In ``module_path``, get every class inheriting from ``class_type``.

    Used by the GUI to dynamically keep track of the :class:`.Loader`,
    :class:`.Model` and :class:`.Plotter` that are implemented.

    Parameters
    ----------
    module_name :
        The name of a module.
    base_class :
        The mother object that classes should inherit from.

    Returns
    -------
        Keys are the name of the objects inheriting from ``base_class`` found
        in ``module_name``. Values are the path leading to them.

    """
    classes: dict[str, str] = {}
    package = __import__(module_name, fromlist=[""])
    for _, name, _ in pkgutil.walk_packages(
        package.__path__, package.__name__ + "."
    ):
        module = __import__(name, fromlist=[""])
        for name, cls in inspect.getmembers(module, inspect.isclass):
            if issubclass(cls, base_class) and cls is not base_class:
                classes[name] = module.__name__
    return classes


def documentation_url(
    obj: Any, *, url_doc_override: str | None = None, **kwargs
) -> str:
    """Infer the link to the API documentation from object path.

    If ``doc_override`` is provided, will return the URL corresponding to this
    path instead.

    """
    if url_doc_override is not None:
        return "/".join((DOC_URL, url_doc_override)) + ".html"
    module = obj.__class__.__module__
    package = module.split(".")[0]
    parts = (DOC_URL, package, module)
    return "/".join(parts) + ".html"


def flatten[T](nest: Iterable[T]) -> Iterator[T]:
    """Flatten nested list of lists of..."""
    for _in in nest:
        if isinstance(_in, Iterable) and not isinstance(_in, (str, bytes)):
            yield from flatten(_in)
        else:
            yield _in



import os
import sys

import eemilib
import sphinx

project = "EEmiLib"
copyright = "2025, Adrien Plaçais"
author = "Adrien Plaçais"
version = eemilib.__version__
release = version

# -- General configuration ---------------------------------------------------
# Add the _ext/ folder so that Sphinx can find it
sys.path.append(os.path.abspath("./_ext"))

extensions = [
    "myst_parser",
    "nbsphinx",
    "sphinx.ext.autodoc",  # include doc from docstrings
    "sphinx.ext.intersphinx",  # interlink with other docs, such as numpy
    "sphinx.ext.napoleon",  # handle numpy style
    "sphinx.ext.todo",  # allow use of TODO
    "sphinx_rtd_theme",  # ReadTheDocs theme
    "sphinxcontrib.bibtex",
    "unit_role",
]

autodoc_default_options = {
    "member-order": "bysource",
    "members": True,
    "private-members": True,
    "special-members": "__init__, __post_init__, __str__",
    "undoc-members": True,
}

add_module_names = False
default_role = "literal"
todo_include_todos = True

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
bibtex_bibfiles = ["references.bib"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_sidebars = {
    "**": [
        "versions.html",
    ],
}

# -- Check that there is no broken link --------------------------------------
nitpicky = True
nitpick_ignore = [
    ("py:class", "PyQt5.QtWidgets.QCheckBox"),
    ("py:class", "PyQt5.QtWidgets.QComboBox"),
    ("py:class", "PyQt5.QtWidgets.QGroupBox"),
    ("py:class", "PyQt5.QtWidgets.QHBoxLayout"),
    ("py:class", "PyQt5.QtWidgets.QLineEdit"),
    ("py:class", "PyQt5.QtWidgets.QListWidget"),
    ("py:class", "PyQt5.QtWidgets.QMainWindow"),
    ("py:class", "PyQt5.QtWidgets.QPushButton"),
    ("py:class", "PyQt5.QtWidgets.QRadioButton"),
    ("py:class", "PyQt5.QtWidgets.QTableWidget"),
    ("py:class", "PyQt5.QtWidgets.QTableWidgetItem"),
    ("py:class", "PyQt5.QtWidgets.QWidget"),
    ("py:class", "T"),
    ("py:class", "optional"),
    ("py:class", "numpy.float64"),
    # Dirty fix
    ("py:class", "ImplementedEmissionData"),
    ("py:class", "ImplementedPop"),
]
intersphinx_mapping = {
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs", None),
    "python": ("https://docs.python.org/3", None),
    "PyQt5": (
        "https://www.riverbankcomputing.com/static/Docs/PyQt5/",
        None,
    ),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}

rst_prolog = """
.. |axplot| replace:: :meth:`matplotlib.axes.Axes.plot`
.. |dfplot| replace:: :meth:`pandas.DataFrame.plot`

"""

# Parameters for sphinx-autodoc-typehints
always_document_param_types = True
always_use_bars_union = True
typehints_defaults = "comma"

# MyST parser to include markdown files
myst_gfm_only = True  # interpret markdown with github styling
#
# -- bug fixes ---------------------------------------------------------------
# fix following warning:
# <unknown>:1: warning: py:class reference target not found: pathlib._local.path [ref.class]
# note that a patch is provided by sphinx 8.2, but nbsphinx 0.9.7 requires
# sphinx<8.2
# associated issue:
# https://github.com/sphinx-doc/sphinx/issues/13178
if sys.version_info[:2] >= (3, 13) and sphinx.version_info[:2] < (8, 2):  # type: ignore
    import pathlib

    from sphinx.util.typing import _INVALID_BUILTIN_CLASSES

    _INVALID_BUILTIN_CLASSES[pathlib.Path] = "pathlib.Path"  # type: ignore
    nitpick_ignore.append(("py:class", "pathlib._local.Path"))



"""Define the ABC :class:`Plotter` to produce the plots."""

from abc import ABC, abstractmethod

import pandas as pd
from eemilib.util.constants import ImplementedPop
from eemilib.util.helper import documentation_url


class Plotter(ABC):
    """A generic object to plot distributions, emission yields, etc."""

    def __init__(self, *args, gui: bool = False, **kwargs) -> None:
        """Instantiate the object.

        Parameters
        ----------
        gui :
            Can be used if using the GUI, eg to activate interactive mode.

        """
        self.doc_url = documentation_url(self)

    @abstractmethod
    def plot_emission_yield[T](
        self,
        df: pd.DataFrame,
        axes: T | None = None,
        population: ImplementedPop | None = None,
        **kwargs,
    ) -> T:
        """Plot emission yield data."""

    @abstractmethod
    def plot_emission_energy_distribution[T](
        self,
        df: pd.DataFrame,
        axes: T | None = None,
        population: ImplementedPop | None = None,
        **kwargs,
    ) -> T:
        """Plot the given emission energy distribution, return Axes object."""

    @abstractmethod
    def plot_emission_angle_distribution[T](
        self,
        df: pd.DataFrame,
        axes: T | None = None,
        population: ImplementedPop | None = None,
        **kwargs,
    ) -> T:
        """Plot the given emission angles distribution, return Axes object."""



"""Define the model related interface in the GUI."""

from eemilib.gui.helper import PARAMETER_ATTR_TO_POS
from PyQt5.QtWidgets import QGroupBox, QTableWidget, QVBoxLayout


def model_configuration() -> tuple[QGroupBox, QTableWidget]:
    """Set the interface related to the model specific parameters."""
    group = QGroupBox("Model configuration")
    layout = QVBoxLayout()

    headers = list(PARAMETER_ATTR_TO_POS.keys())
    n_cols = len(headers)
    model_table = QTableWidget(0, n_cols)
    model_table.setHorizontalHeaderLabels(headers)
    model_table.setMaximumHeight(1000)
    model_table.setMinimumHeight(200)
    layout.addWidget(model_table)

    group.setLayout(layout)
    return group, model_table



"""Define some types and constants."""

from typing import Literal

q = 1.6e-19

#: The three types of emission data.
ImplementedEmissionData = Literal[
    "Emission Yield", "Emission Energy", "Emission Angle"
]
IMPLEMENTED_EMISSION_DATA = (
    "Emission Yield",
    "Emission Energy",
    "Emission Angle",
)

#: Implemented populations.
ImplementedPop = Literal["SE", "EBE", "IBE", "all"]
IMPLEMENTED_POP = ("SE", "EBE", "IBE", "all")

#: Typical column where energy is stored in dataframes.
#: In emission yield data, this is the energy of PEs. In emission energy data,
#: this is the energy of emitted electrons.
col_energy = "Energy [eV]"
#: Typical column where normal incidence data is stored in dataframes.
col_normal = "0.0 [deg]"

md_ey: dict[ImplementedPop, str] = {
    "SE": r"SEEY $\delta$",
    "EBE": r"EBEEY $\eta_e$",
    "IBE": r"IBEEY $\eta_i$",
    "all": r"TEEY $\sigma$",
}
md_energy_distrib: dict[ImplementedPop, str] = {
    "SE": r"$f_\mathrm{SE}$",
    "EBE": r"$f_\mathrm{EBE}$",
    "IBE": r"$f_\mathrm{IBE}$",
    "all": r"$f_\mathrm{all}$",
}
#: y-label in plots for the different emission data type.
md_ylabel: dict[ImplementedEmissionData, str] = {
    "Emission Yield": "Emission Yield",
    "Emission Energy": r"Emission Energy Distribution [$\mathrm{eV}^{-1}$]",
    "Emission Angle": "Emission Angle Distribution",
}



"""Define the file selection matrix."""

from collections.abc import Callable

from eemilib.util.constants import IMPLEMENTED_EMISSION_DATA, IMPLEMENTED_POP
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import (
    QFileDialog,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QMainWindow,
    QPushButton,
)


def file_selection_matrix(
    main_window: QMainWindow,
) -> tuple[QGroupBox, list[list[None | QListWidget]]]:
    """Create the 4 * 3 matrix to select the files to load."""
    file_matrix_group = QGroupBox("Files selection matrix")
    file_matrix_layout = QGridLayout()

    row_labels, col_labels = IMPLEMENTED_POP, IMPLEMENTED_EMISSION_DATA
    n_rows, n_cols = len(row_labels), len(col_labels)

    file_lists: list[list[None | QListWidget]]
    file_lists = [[None for _ in range(n_cols)] for _ in range(n_rows)]

    for i, label in enumerate(row_labels):
        file_matrix_layout.addWidget(QLabel(label), i + 1, 0)

    for j, label in enumerate(col_labels):
        file_matrix_layout.addWidget(QLabel(label), 0, j + 1)

    for i in range(n_rows):
        for j in range(n_cols):
            cell_layout = QHBoxLayout()
            button, file_list = _setup_file_selection_widget(
                lambda _, x=i, y=j: _select_files(
                    main_window, file_lists, x, y
                )
            )
            cell_layout.addWidget(button)
            cell_layout.addWidget(file_list)
            file_lists[i][j] = file_list

            file_matrix_layout.addLayout(cell_layout, i + 1, j + 1)

    file_matrix_group.setLayout(file_matrix_layout)
    return file_matrix_group, file_lists


def _setup_file_selection_widget(
    select_file_func: Callable,
) -> tuple[QPushButton, QListWidget]:
    """Set the button to load and the list of selected files."""
    button = QPushButton("📂")
    button.setFont(QFont("Segoe UI Emoji", 10))
    button.clicked.connect(select_file_func)

    file_list = QListWidget()
    return button, file_list


def _select_files(
    main_window: QMainWindow,
    files_list: list[list[None | QListWidget]],
    row: int,
    col: int,
) -> None:
    """Set up a function to set the filepaths."""
    options = QFileDialog.Options()
    file_names, _ = QFileDialog.getOpenFileNames(
        main_window,
        "Select Files",
        "",
        "All Files (*);;CSV Files (*.csv)",
        options=options,
    )
    if file_names:
        current_file_lists = files_list[row][col]
        assert current_file_lists is not None
        current_file_lists.clear()
        current_file_lists.addItems(file_names)
        # self.data_matrix.set_files(file_names, row=row, col=col)



"""Define functions to be as DRY as possible."""

import logging
from abc import ABCMeta
from collections.abc import Collection
from functools import partial
from typing import Any, Literal, overload

from eemilib.model.parameter import Parameter
from eemilib.util.helper import get_classes
from PyQt5.QtCore import Qt, QUrl
from PyQt5.QtGui import QDesktopServices, QDoubleValidator, QIntValidator
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QRadioButton,
    QWidget,
)


def setup_dropdown(
    module_name: str,
    base_class: ABCMeta,
    buttons_args: dict[str, Any],
) -> tuple[dict[str, str], QHBoxLayout, QComboBox, list[QPushButton]]:
    """Set up interface with a dropdown menu and a button next to it.

    Parameters
    ----------
    module_name :
        Where the entries of the dropdown will be searched.
    base_class :
        The base class from which dropdown entries should inherit.
    buttons :
        Dictionary where the keys are the name of the buttons to add next to
        the dropdown menu, and values the callable that will be called when
        clicking the button.

    Returns
    -------
    dict[str, str]
        Keys are the name of the objects inheriting from ``base_class`` found
        in ``module_name``. Values are the path leading to them.
    QHBoxLayout
        Layout holding together ``dropdown`` and ``button``.
    QComboBox
        Dropdown menu holding the keys of ``classes``.
    list[QPushButton]
        The buttons next to the dropdown menu.

    """
    classes = get_classes(module_name, base_class)

    layout = QHBoxLayout()

    dropdown = QComboBox()
    dropdown.addItems(classes.keys())
    layout.addWidget(QLabel(f"Select {base_class.__name__}:"))
    layout.addWidget(dropdown)

    buttons = []
    for name, action in buttons_args.items():
        button = QPushButton(name)
        button.clicked.connect(action)
        layout.addWidget(button)
        buttons.append(button)

    return classes, layout, dropdown, buttons


def set_dropdown_value(
    dropdown: QComboBox, value: str | ABCMeta | None
) -> None:
    """Set a ``dropdown`` to desired value.

    Parameters
    ----------
    dropdown :
        Dropdown object.
    value :
        Name of class or class object you want to select in the dropdown. If
        unset, we do not do anything.
    allowed_values :
        Dict used for the ``dropdown`` creation; links name of class objects
        to their import path.

    """
    if value is None:
        return
    if isinstance(value, ABCMeta):
        value = value.__name__
    index = dropdown.findText(value)
    if index == -1:
        logging.info(f"{value = } not found in {dropdown = } items.")
        return
    dropdown.setCurrentIndex(index)


def setup_linspace_entries(
    label: str,
    initial_values: tuple[float, float, int],
    max_value: float | None = None,
) -> tuple[QHBoxLayout, QLineEdit, QLineEdit, QLineEdit]:
    """Create an input to call np.linspace."""
    layout = QHBoxLayout()
    layout.addWidget(QLabel(label))

    widgets: list[QWidget] = []
    for label, is_int, x_0, x_max in zip(
        ("first", "last", "n_points"),
        (False, False, True),
        initial_values,
        (max_value, max_value, None),
    ):
        layout.addWidget(QLabel(label))
        widgets.append(w := _linspace_entry(is_int, x_0=x_0, x_max=x_max))
        layout.addWidget(w)

    return layout, widgets[0], widgets[1], widgets[2]


def _linspace_entry(
    is_int: bool, x_0: float, x_min: int = 0, x_max: float | None = None
) -> QWidget:
    """Create widget for a single linspace entry."""
    validator = QDoubleValidator()
    validator.setBottom(x_min)
    if is_int:
        validator = QIntValidator()
    if x_max is not None:
        validator.setTop(int(x_max))

    entry = QLineEdit()
    entry.setValidator(validator)
    entry.setText(str(x_0))
    return entry


def setup_lock_checkbox(
    parameter: Parameter,
) -> QWidget:
    """Create the checkbox for the Lock button."""
    checkbox = QCheckBox()
    checkbox.setChecked(parameter.is_locked)
    checkbox.stateChanged.connect(
        lambda state, param=parameter: _toggle_lock(state, param)
    )

    checkbox_widget = QWidget()
    layout = QHBoxLayout(checkbox_widget)
    layout.addWidget(checkbox)
    layout.setAlignment(Qt.AlignCenter)
    layout.setContentsMargins(0, 0, 0, 0)
    checkbox_widget.setLayout(layout)
    return checkbox_widget


def _toggle_lock(state: Any, parameter: Parameter) -> None:
    """Activate/deactivate lock."""
    if state == Qt.Checked:
        parameter.lock()
    parameter.unlock()


@overload
def to_plot_checkboxes(
    label: str,
    boxes_labels: Collection[str],
    *,
    several_can_be_checked: Literal[False],
) -> tuple[QHBoxLayout, list[QRadioButton]]: ...


@overload
def to_plot_checkboxes(
    label: str,
    boxes_labels: Collection[str],
    *,
    several_can_be_checked: Literal[True],
) -> tuple[QHBoxLayout, list[QCheckBox]]: ...


def to_plot_checkboxes(
    label: str,
    boxes_labels: Collection[str],
    *,
    several_can_be_checked: bool = False,
) -> tuple[QHBoxLayout, list[QRadioButton] | list[QCheckBox]]:
    """Create several check boxes next to each other."""
    checkbox_constructor = QCheckBox
    if not several_can_be_checked:
        checkbox_constructor = QRadioButton
    checkboxes = [checkbox_constructor(x) for x in boxes_labels]

    layout = QHBoxLayout()
    layout.addWidget(QLabel(label))
    for checkbox in checkboxes:
        layout.addWidget(checkbox)

    return layout, checkboxes


def set_help_button_action(button: QPushButton, obj: Any) -> None:
    """Update the link of the provided help button."""
    button.clicked.disconnect()
    this_help = partial(_open_help, obj=obj)
    button.clicked.connect(this_help)


def _open_help(obj: Any) -> None:
    """Open the ``doc_url`` attribute of given object."""
    url = getattr(obj, "doc_url", None)
    if not isinstance(url, str):
        logging.warning(f"No valid URL found for {obj = }")
        return
    QDesktopServices.openUrl(QUrl(url))


def format_number(value: int | float) -> str:
    """Format the given number.

    Parameters
    ----------
    value :
        Number to format.

    """
    if isinstance(value, int):
        return str(value)
    if isinstance(value, float):
        return f"{value:.2f}"
    return str(value)


# Associate Parameters attributes with their column position
# Note that "name" is the key in the Model.parameters dict rather than the
# Parameter.name attribute (which is not consistent)
PARAMETER_ATTR_TO_POS = {
    "name": 0,
    "unit": 1,
    "value": 2,
    "lower_bound": 3,
    "upper_bound": 4,
    "lock": 5,
}

#: Maps column position in list of parameters to the corresponding Parameter
#: attribute
PARAMETER_POS_TO_ATTR = {
    val: key for key, val in PARAMETER_ATTR_TO_POS.items()
}



"""Store the filepaths entered by user.

.. todo::
    Methods to reset filepaths/data

"""

import logging
from collections.abc import Collection
from pathlib import Path
from typing import Literal, overload

from eemilib.core.model_config import ModelConfig
from eemilib.emission_data.emission_angle_distribution import (
    EmissionAngleDistribution,
)
from eemilib.emission_data.emission_data import EmissionData
from eemilib.emission_data.emission_energy_distribution import (
    EmissionEnergyDistribution,
)
from eemilib.emission_data.emission_yield import EmissionYield
from eemilib.loader.loader import Loader
from eemilib.plotter.plotter import Plotter
from eemilib.util.constants import (
    IMPLEMENTED_EMISSION_DATA,
    IMPLEMENTED_POP,
    ImplementedEmissionData,
    ImplementedPop,
)
from eemilib.util.helper import flatten

pop_to_row = {pop: i for i, pop in enumerate(IMPLEMENTED_POP)}
row_to_pop = {val: key for key, val in pop_to_row.items()}

emission_data_type_to_col = {
    data_type: j for j, data_type in enumerate(IMPLEMENTED_EMISSION_DATA)
}
col_to_emission_data_type = {
    val: key for key, val in emission_data_type_to_col.items()
}

n_rows = len(IMPLEMENTED_POP)
n_cols = len(IMPLEMENTED_EMISSION_DATA)


class DataMatrix:
    """Store all the input files and corresp data in a single object."""

    def __init__(self) -> None:
        """Instantiate the object."""
        self.files_matrix: list[
            list[None | str | Collection[str] | Path | Collection[Path]]
        ]
        self.files_matrix = [
            [None for _ in range(n_cols)] for _ in range(n_rows)
        ]

        self.data_matrix: list[
            list[None | EmissionData | Collection[EmissionData]]
        ]
        self.data_matrix = [
            [None for _ in range(n_cols)] for _ in range(n_rows)
        ]

    def _natures_to_indexes(
        self,
        population_type: ImplementedPop,
        emission_data_type: ImplementedEmissionData,
    ) -> tuple[int, int]:
        """Give the desired indexes."""
        return (
            pop_to_row[population_type],
            emission_data_type_to_col[emission_data_type],
        )

    def _indexes_to_natures(
        self, row: int, col: int
    ) -> tuple[ImplementedPop, ImplementedEmissionData]:
        """Give the desired natures."""
        population_type = row_to_pop[row]
        assert population_type in IMPLEMENTED_POP
        emission_data_type = col_to_emission_data_type[col]
        assert emission_data_type in IMPLEMENTED_EMISSION_DATA
        return population_type, emission_data_type

    @overload
    def set_files(
        self,
        files: str | Path | Collection[str] | Collection[Path],
        row: int,
        col: int,
        population: None,
        emission_data_type: None,
    ) -> None: ...

    @overload
    def set_files(
        self,
        files: str | Path | Collection[str] | Collection[Path],
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: ImplementedEmissionData,
    ) -> None: ...

    def set_files(
        self,
        files: str | Path | Collection[str] | Collection[Path],
        row: int | None = None,
        col: int | None = None,
        population: ImplementedPop | None = None,
        emission_data_type: ImplementedEmissionData | None = None,
    ) -> None:
        """Set the file(s) by index or name."""
        if population and emission_data_type:
            row, col = self._natures_to_indexes(
                population_type=population,
                emission_data_type=emission_data_type,
            )

        if row is None or col is None:
            raise ValueError(
                "You need to provide row and col, or population and "
                f"emission_data_type.\n{row = }, {col = }, {population = },"
                f"{emission_data_type = }"
            )

        self.files_matrix[row][col] = files

    @overload
    def set_data(
        self,
        emission_data: EmissionData | Collection[EmissionData],
        row: int,
        col: int,
        population: None,
        emission_data_type: None,
    ) -> None: ...

    @overload
    def set_data(
        self,
        emission_data: EmissionYield | Collection[EmissionYield],
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: Literal["Emission Yield"],
    ) -> None: ...

    @overload
    def set_data(
        self,
        emission_data: (
            EmissionEnergyDistribution | Collection[EmissionEnergyDistribution]
        ),
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: Literal["Emission Energy"],
    ) -> None: ...

    @overload
    def set_data(
        self,
        emission_data: (
            EmissionAngleDistribution | Collection[EmissionAngleDistribution]
        ),
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: Literal["Emission Angle"],
    ) -> None: ...

    def set_data(
        self,
        emission_data: EmissionData | Collection[EmissionData],
        row: int | None = None,
        col: int | None = None,
        population: ImplementedPop | None = None,
        emission_data_type: ImplementedEmissionData | None = None,
    ) -> None:
        """Set the data by index or name."""
        if population and emission_data_type:
            row, col = self._natures_to_indexes(
                population_type=population,
                emission_data_type=emission_data_type,
            )

        if row is None or col is None:
            raise ValueError(
                "You need to provide row and col, or population and "
                f"emission_data_type.\n{row = }, {col = }, {population = },"
                f"{emission_data_type = }"
            )

        self.data_matrix[row][col] = emission_data

    @overload
    def get_files(
        self,
        row: int,
        col: int,
        population: None,
        emission_data_type: None,
    ) -> None | str | Path | Collection[str] | Collection[Path]: ...

    @overload
    def get_files(
        self,
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: ImplementedEmissionData,
    ) -> None | str | Path | Collection[str] | Collection[Path]: ...

    def get_files(
        self,
        row: int | None = None,
        col: int | None = None,
        population: ImplementedPop | None = None,
        emission_data_type: ImplementedEmissionData | None = None,
    ) -> None | str | Path | Collection[str] | Collection[Path]:
        """Get the file(s) by index or name."""
        if population and emission_data_type:
            row, col = self._natures_to_indexes(
                population_type=population,
                emission_data_type=emission_data_type,
            )

        if row is None or col is None:
            raise ValueError(
                "You need to provide row and col, or population and "
                f"emission_data_type.\n{row = }, {col = }, {population = },"
                f"{emission_data_type = }"
            )

        return self.files_matrix[row][col]

    @overload
    def get_data(
        self,
        row: int,
        col: int,
        population: None,
        emission_data_type: None,
    ) -> None | EmissionData | Collection[EmissionData]: ...

    @overload
    def get_data(
        self,
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: Literal["Emission Yield"],
    ) -> None | EmissionYield | Collection[EmissionYield]: ...

    @overload
    def get_data(
        self,
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: Literal["Emission Energy"],
    ) -> (
        None
        | EmissionEnergyDistribution
        | Collection[EmissionEnergyDistribution]
    ): ...

    @overload
    def get_data(
        self,
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: Literal["Emission Angle"],
    ) -> (
        None
        | EmissionAngleDistribution
        | Collection[EmissionAngleDistribution]
    ): ...

    @overload
    def get_data(
        self,
        row: None,
        col: None,
        population: None,
        emission_data_type: None,
    ) -> Collection[EmissionData]: ...

    @overload
    def get_data(
        self,
        row: None,
        col: None,
        population: ImplementedPop,
        emission_data_type: None,
    ) -> Collection[EmissionData]: ...
    @overload
    def get_data(
        self,
        row: None,
        col: None,
        population: None,
        emission_data_type: ImplementedEmissionData,
    ) -> Collection[EmissionData]: ...

    def get_data(
        self,
        row: int | None = None,
        col: int | None = None,
        population: ImplementedPop | None = None,
        emission_data_type: ImplementedEmissionData | None = None,
    ) -> None | EmissionData | Collection[EmissionData]:
        """Get the file(s) by index or name.

        You can provide ``row`` and ``col`` directly.

        Alternatively, provide ``population`` and ``emission_data_type``. If
        ``population`` is not given, return valid data corresponding to all
        populations. If ``emission_data_type`` is not given, return valid
        data corresponding to all emission data.

        Parameters
        ----------
        row :
            Row index in data matrix.
        col :
            Column index in data matrix.
        population :
            Population type.
        emission_data_type :
            Emission data type.

        Returns
        -------
            Desired data; if the specified data does not exists, a ``None`` is
            returned without any error message.

        """
        if population and emission_data_type:
            row, col = self._natures_to_indexes(
                population_type=population,
                emission_data_type=emission_data_type,
            )

        if population and emission_data_type is None:
            single_pop_data = [
                self.get_data(
                    population=population, emission_data_type=data_type
                )
                for data_type in IMPLEMENTED_EMISSION_DATA
            ]
            return [d for d in flatten(single_pop_data) if d is not None]

        if emission_data_type and population is None:
            emission_data = [
                self.get_data(
                    population=pop, emission_data_type=emission_data_type
                )
                for pop in IMPLEMENTED_POP
            ]
            return [d for d in flatten(emission_data) if d is not None]

        if row is None and col is None:
            all_data = [
                [
                    self.get_data(population=pop, emission_data_type=data_type)
                    for pop in IMPLEMENTED_POP
                ]
                for data_type in IMPLEMENTED_EMISSION_DATA
            ]
            return [d for d in flatten(all_data) if d is not None]

        if row is None or col is None:
            raise ValueError(
                "You need to provide row and col, or population and "
                f"emission_data_type.\n{row = }, {col = }, {population = },"
                f"{emission_data_type = }"
            )

        return self.data_matrix[row][col]

    def load_data(self, loader: Loader) -> None:
        """Load all filepaths in ``files_matrix``.

        .. todo::
            Could be more concise.

        """
        for pop in IMPLEMENTED_POP:
            for data_type in IMPLEMENTED_EMISSION_DATA:
                filepath = self.get_files(
                    population=pop, emission_data_type=data_type
                )  # type: ignore

                if not filepath:
                    continue

                emission_data = None
                if data_type == "Emission Yield":
                    emission_data = EmissionYield.from_filepath(
                        pop, loader, *filepath
                    )

                elif data_type == "Emission Energy":
                    emission_data = EmissionEnergyDistribution.from_filepath(
                        pop, loader, *filepath
                    )

                elif data_type == "Emission Angle":
                    emission_data = EmissionAngleDistribution.from_filepath(
                        pop, loader, *filepath
                    )

                if emission_data:
                    self.set_data(
                        emission_data,
                        population=pop,
                        emission_data_type=data_type,
                    )  # type: ignore

    def has_all_mandatory_files(self, model_config: ModelConfig) -> bool:
        """Tell if files defined by :attr:`.Model.model_config` are set."""
        for emission_data_type, corresponding_attribute in zip(
            IMPLEMENTED_EMISSION_DATA,
            (
                "emission_yield_files",
                "emission_energy_files",
                "emission_angle_files",
            ),
        ):
            mandatory_populations = getattr(
                model_config, corresponding_attribute
            )

            for mandatory_population in mandatory_populations:
                if mandatory_population not in IMPLEMENTED_POP:
                    logging.error(
                        f"{mandatory_population = } not in "
                        f"{IMPLEMENTED_POP = }"
                    )
                    return False

                filepath = self.get_files(
                    population=mandatory_population,
                    emission_data_type=emission_data_type,
                )  # type: ignore
                if filepath is None:
                    logging.error(
                        f"You must define a {emission_data_type} filepath for"
                        f" population {mandatory_population}"
                    )
                    return False

                data_object = self.get_data(
                    population=mandatory_population,
                    emission_data_type=emission_data_type,
                )  # type: ignore
                if data_object is None:
                    logging.error(
                        f"You must load {emission_data_type} filepath for "
                        f"population {mandatory_population}"
                    )
                    return False
        return True

    def plot[T](
        self,
        plotter: Plotter,
        population: ImplementedPop | Collection[ImplementedPop],
        emission_data_type: ImplementedEmissionData,
        axes: T | None = None,
        **kwargs,
    ) -> T | None:
        """Plot desired measured data using ``plotter``.

        This method uses :meth:`.DataMatrix.get_data` to get the
        :class:`.EmissionData` instance matching ``population`` and
        ``emission_data_type``. Then it calls the :meth:`.EmissionData.plot`
        method.

        Parameters
        ----------
        plotter :
            Object realizing the plot. We transfer it to the
            :meth:`.EmissionData.plot` method.
        population :
            One or several populations to plot. If several are given, we simply
            recursively call this method.
        emission_data_type :
            Type of data to plot.
        axes :
            Axes to re-use if given.
        kwargs :
            Other keyword arguments passed to the :meth:`.EmissionData.plot`
            method.

        Returns
        -------
            Created axes object, or ``None`` if no plot was created.

        """
        if isinstance(population, Collection) and not isinstance(
            population, str
        ):
            for pop in population:
                axes = self.plot(
                    plotter, pop, emission_data_type, axes=axes, **kwargs
                )
            return axes

        emission_data = self.get_data(
            population=population, emission_data_type=emission_data_type
        )

        if emission_data is None:
            logging.info(
                f"No measurement found for {population = } and "
                f"{emission_data_type = }. Skipping this plot."
            )
            return axes

        if isinstance(emission_data, EmissionData):
            emission_data = (emission_data,)

        for data in emission_data:
            axes = data.plot(
                plotter,
                axes=axes,
                population=population,
                **kwargs,
            )
        return axes

    @property
    def teey(self) -> EmissionYield:
        """Return the TEEY directly."""
        emission_yield = self.data_matrix[3][0]
        assert isinstance(
            emission_yield, EmissionYield
        ), f"Incorrect type for emission_yield: {type(emission_yield)}"
        assert emission_yield.population == "all"
        return emission_yield

    @property
    def se_energy_distribution(self) -> EmissionEnergyDistribution:
        """Return the energy distribution of SEs."""
        distrib = self.data_matrix[0][1]
        assert isinstance(
            distrib, EmissionEnergyDistribution
        ), f"Incorrect type for energy distribution: {type(distrib)}"
        assert distrib.population == "SE"
        return distrib

    @property
    def all_energy_distribution(self) -> EmissionEnergyDistribution:
        """Return the energy distribution of all emitted electrons."""
        distrib = self.data_matrix[3][1]
        assert isinstance(
            distrib, EmissionEnergyDistribution
        ), f"Incorrect type for energy distribution: {type(distrib)}"
        assert distrib.population == "all"
        return distrib



"""Define some common helpers for loading data."""

from pathlib import Path

from eemilib.util.constants import col_energy


def read_header(
    filepath: str | Path,
    sep: str = "\t",
    comment: str = "#",
) -> tuple[list[str], int]:
    """Get the line describing columns content.

    It is the first line of the files that does not start with a comment
    character. Header of first column can be anything. Header of following
    columns must hold incidence angle and be convertable to a float.

    Parameters
    ----------
    filepath :
        Path to file holding data under study.
    sep :
        Column delimiter.
    comment :
        Comment character.

    Returns
    -------
    list[str]
        Columns descriptors. First column is ``Energy [eV]``. Following is/are
        ``theta [deg]``, where ``theta`` is the value of the incidence angle.
    int
        Number of comment lines before the header.

    """
    header = []
    n_comments = 0
    with open(filepath) as file:
        for n_comments, line in enumerate(file):
            if not line.startswith(comment):
                header = line.strip().split(sep)
                break
    if not header:
        raise OSError(
            f"Error reading {filepath}. It seems there is no uncommented line?"
            f"Comment character is {comment}."
        )

    return _format_header(header), n_comments


def _format_header(header: list[str]) -> list[str]:
    """Generate default header."""
    header[0] = col_energy
    header[1:] = [f"{float(h)} [deg]" for h in header[1:]]
    return header


def read_comments(filepath: str | Path, comment: str = "#") -> list[str]:
    """Read the comments in the file.

    Parameters
    ----------
    filepath :
        Path to file holding data under study.
    comment :
        Comment character.

    Returns
    -------
    list[str]
        Comments, line by line. Without the comment character.

    """
    comments: list[str] = []
    with open(filepath) as file:
        for line in file:
            if not line.startswith(comment):
                return comments
            comments.append(line[1:])
    return comments



"""Define objects to load the different formats of electron emission files."""

from .deesse_loader import DeesseLoader
from .pandas_loader import PandasLoader

__all__ = ["DeesseLoader", "PandasLoader"]



"""Define functions to extract some characteristics from emission data."""

import numpy as np
import pandas as pd
from eemilib.util.constants import col_energy, col_normal


def trim(
    normal_ey: pd.DataFrame,
    min_e: float = -1.0,
    max_e: float = -1.0,
) -> pd.DataFrame:
    """Remove EY outside of given energy range (if provided).

    Parameters
    ----------
    normal_ey :
        Holds normal emission yield. Columns are ``EY_col1`` (energy, stored
        by increasing values) and ``EY_colnorm`` (normal EY).
    min_e :
        Energy at which the output dataframe should start (if provided). The
        default is a negative value, in which case the output dataframe is not
        bottom-trimed.
    max_e :
        Energy at which the output dataframe should end (if provided). The
        default is a negative value, in which case the output dataframe is not
        top-trimed.

    Returns
    -------
        ``normal_ey`` but with energies ranging only from ``min_e`` to
        ``max_e``.

    """
    if min_e >= 0:
        trimed = normal_ey[normal_ey[col_energy] >= min_e]
        assert isinstance(trimed, pd.DataFrame)
        normal_ey = trimed
    if max_e >= 0:
        trimed = normal_ey[normal_ey[col_energy] <= max_e]
        assert isinstance(trimed, pd.DataFrame)
        normal_ey = trimed

    return normal_ey.reset_index(drop=True)


def resample(ey: pd.DataFrame, n_interp: int = -1) -> pd.DataFrame:
    """Return the emission yield with more points and/or updated limits."""
    if n_interp < 0:
        return ey
    new_ey = {
        col_energy: np.linspace(
            ey[col_energy].min(), ey[col_energy].max(), n_interp
        )
    }
    for col_name in ey.columns:
        if col_name == col_energy:
            continue
        new_ey[col_name] = np.interp(
            x=new_ey[col_energy],
            xp=ey[col_energy],
            fp=ey[col_name],
        )

    return pd.DataFrame(new_ey)


def get_emax_eymax(normal_ey: pd.DataFrame) -> tuple[float, float]:
    """Get energy and max emission yields."""
    ser_max = normal_ey.loc[normal_ey[col_normal].idxmax()]
    e_max = ser_max[col_energy]
    ey_max = ser_max[col_normal]
    return e_max, ey_max


def get_crossover_energies(
    normal_ey: pd.DataFrame, e_max: float, min_e: float = 10.0
) -> tuple[tuple[float, float], tuple[float, float]]:
    """Compute first and second crossover energies.

    Parameters
    ----------
    normal_ey :
        Holds energy of PEs as well as emission yield at nominal incidence.
    e_max :
        Energy of maximum emission yield. Used to discriminate
        :math:`E_{c1}` from :math:`E_{c2}`.
    min_e :
        Energy under which :math:`E_{c1}` is not searched. It is useful if
        emission yield data comes from a model which sets the emission
        yield to unity at very low energies (eg some implementations of
        Vaughan). The default value is 10 eV.

    Returns
    -------
    tuple[float, float]
        First crossover energy, corresponding emission yield.
    tuple[float, float]
        Second crossover energy, corresponding emission yield.

    """
    first_half = trim(normal_ey, min_e=min_e, max_e=e_max)
    ser_ec1 = first_half.loc[(first_half[col_normal] - 1.0).abs().idxmin()]
    ec1 = ser_ec1[col_energy]
    ey_ec1 = ser_ec1[col_normal]

    second_half = trim(normal_ey, min_e=e_max)
    ser_ec2 = second_half.loc[(second_half[col_normal] - 1.0).abs().idxmin()]
    ec2 = ser_ec2[col_energy]
    ey_ec2 = ser_ec2[col_normal]

    return (ec1, ey_ec1), (ec2, ey_ec2)


def get_ec1(
    normal_ey: pd.DataFrame,
    min_e: float = -1.0,
    max_e: float = -1.0,
    n_interp: int = -1,
    **kwargs,
) -> float:
    """Interpolate the energy vs teey array and give the E_c1."""
    if min_e < 0.0:
        min_e = np.nanmin()
    ene_interp = np.linspace(0.0, 500.0, 10001)

    # Whith Vaughan, and with seey_low = 1, avoid detecting ec1 below E0
    if min_e is not None:
        ene_interp = np.linspace(min_e + 1.0, 500.0, 1001)

    teey_interp = np.interp(ene_interp, ey[:, 0], ey[:, 1], left=0.0)
    idx = np.argmin(np.abs(teey_interp - 1.0))
    ec1 = ene_interp[idx]
    return ec1


def get_max(teey: np.ndarray, E0: float = None, **kwargs) -> (float, float):
    """Interpolate the energy vs teey array and give the E and sigma max."""
    ene_interp = np.linspace(0.0, 1e3, 10001)

    # Whith Vaughan, and with seey_low = 1, avoid detecting ec1 below E0
    if E0 is not None:
        ene_interp = np.linspace(E0 + 1.0, 1e3, 1001)

    teey_interp = np.interp(ene_interp, teey[:, 0], teey[:, 1], left=0.0)
    idx = np.argmax(teey_interp)
    return ene_interp[idx], teey_interp[idx]



"""Define a generic files loader.

Check the documentation of :meth:`.PandasLoader.load_emission_yield` and
:meth:`.PandasLoader.load_emission_energy_distribution` for expected file
formats.

"""

import logging
from pathlib import Path
from typing import Any

import pandas as pd
from eemilib.loader.helper import read_comments, read_header
from eemilib.loader.loader import Loader


class PandasLoader(Loader):
    """Define the pandas loader."""

    def __init__(self) -> None:
        """Init object."""
        return super().__init__()

    def load_emission_yield(
        self,
        filepath: str | Path,
        sep: str = ",",
        comment: str = "#",
    ) -> pd.DataFrame:
        """Load and format the given emission yield file.

        ``CSV`` files can have comments at the start of the file, starting with
        a ``#`` character. Column separator must be ``,``. First non-commented
        line is incidence angle in degrees. First column is incident energy in
        :unit:`eV`. EY is the next columns (excluding the first line).
        Example:

        .. code-block::

            # Cu measurements
            # Some comments
            # Energy | 0deg | 20deg | 40deg | 60deg
            0,0,20,40,60
            0,0.814,0.781,0.866,0.918
            10,0.574,0.553,0.637,0.803
            20,0.632,0.594,0.671,0.817

        Files in the :file:`data/example_copper/` files are taken from
        :cite:`Placais2020b` and are correctly formatted.

        Parameters
        ----------
        filepath :
            Path to file holding data under study.
        sep :
            Column delimiter.
        comment :
            Comment character.

        Returns
        -------
        pandas.DataFrame
            Structure holding the data. Has a ``Energy [eV]`` column
            holding PEs energy. And one or several columns ``theta [deg]``,
            where ``theta`` is the value of the incidence angle and content is
            corresponding emission yield.

        """
        header, n_comments = read_header(filepath, sep, comment)
        df = pd.read_csv(
            filepath,
            comment=comment,
            sep=sep,
            names=header,
            skiprows=n_comments + 1,
        )
        logging.info(f"Successfully loaded emission yield file(s) {filepath}")
        return df

    def load_emission_angle_distribution(self, *args) -> Any:
        raise NotImplementedError

    def load_emission_energy_distribution(
        self,
        filepath: str | Path,
        sep: str = ",",
        comment: str = "#",
    ) -> tuple[pd.DataFrame, float | None]:
        """Load and format the given emission energy file.

        ``CSV`` files can have comments at the start of the file, starting with
        a ``#`` character. It is expected that the energy of PEs used for the
        measurements is on the second commented line, in :unit:`eV`. Column
        separator must be ``,``. First non-commented line is incidence angle in
        degrees. First column is emission energy in :unit:`eV`. Distribution is
        in the next columns (excluding the first line).
        Example:

        .. code-block::

            # PEs energy in eV
            # 100
            0,0
            1.999999999999975131e-01,7.117578753770542783e-03
            3.999999999999968026e-01,1.138131444290255145e-02
            5.999999999999978684e-01,1.510969903349285159e-02

        Files in the :file:`data/example_ag/emission_energy` files are
        correctly formatted.

        Parameters
        ----------
        filepath :
            Path to file holding data under study.
        sep :
            Column delimiter.
        comment :
            Comment character.

        Returns
        -------
        pd.DataFrame
            Structure holding the data. Has a ``Energy [eV]`` column
            holding emitted electrons energy. And one or several columns
            ``theta [deg]``, where ``theta`` is the value of the incidence
            angle and content is corresponding emission energy distribution.
        float
            Energy of Primary Electrons in :unit:`eV`. If not found in the file
            comments, it will be inferred from the position of the EBEs peak.

        """
        header, n_comments = read_header(filepath, sep, comment)
        df = pd.read_csv(
            filepath,
            comment=comment,
            sep=sep,
            names=header,
            skiprows=n_comments + 1,
        )
        if len(df.columns) != 2:
            raise RuntimeError(
                f"Error loading {filepath}. "
                f"The file should have two columns, separated by a ``{sep}`` "
                f"character. File was read as:\n{df}"
            )

        comments = read_comments(filepath, comment=comment)

        if len(comments) < 2:
            logging.error(
                f"Error loading {filepath}. "
                "PandasLoader expects at least two lines of comments at the "
                "start of filepath. (Second line should hold energy of primary"
                "electrons in eV). Will try to infer this quantity from the "
                "position of EBEs peak."
            )
            return df, None

        try:
            e_pe = float(comments[1])

        except ValueError as e:
            logging.error(
                f"Error loading {filepath}. "
                "PandasLoader expects the second comment line to hold the "
                "energy of PEs, in eV. Will try to infer this quantity "
                f"from the position of EBEs peak.\n{e}"
            )
            return df, None

        logging.info(
            "Successfully loaded emission energy distribution file(s) "
            f"{filepath}"
        )
        return df, e_pe



r"""Create the Vaughan model, to compute TEEY.

TEEY at non-normal incidence will not be taken into account into the fit
(FIXME).

.. todo::
    Make this more robust. Especially the _vaughan_func.

"""

import logging
import math
from typing import Any, Literal, TypedDict

import numpy as np
import pandas as pd
from eemilib.core.model_config import ModelConfig
from eemilib.emission_data.data_matrix import DataMatrix
from eemilib.model.model import Model
from eemilib.model.parameter import Parameter
from eemilib.util.constants import ImplementedEmissionData, ImplementedPop
from numpy.typing import NDArray
from scipy.optimize import least_squares

VaughanImplementation = Literal["original", "CST", "SPARK3D"]
E_0_SPARK3D = 10.0


class VaughanParameters(TypedDict):
    E_0: Parameter
    E_max: Parameter
    delta_E_transition: Parameter
    teey_low: Parameter
    teey_max: Parameter
    k_s: Parameter
    k_se: Parameter
    E_c1: Parameter


class Vaughan(Model):
    """Define the classic Vaughan model."""

    emission_data_types = ["Emission Yield"]
    populations = ["all"]
    considers_energy = True
    is_3d = True
    is_dielectrics_compatible = False
    model_config = ModelConfig(
        emission_yield_files=("all",),
        emission_energy_files=(),
        emission_angle_files=(),
    )
    initial_parameters = {
        "E_0": {
            "markdown": r"E_0",
            "unit": "eV",
            "value": 12.5,
            "description": r"Threshold energy. By default, locked to "
            + r":math:`12.5\mathrm{\,eV}`. If unlocked, will be fitted to "
            + r"retrieve :math:`E_{c,\,1}`.",
            "is_locked": True,
        },
        "E_max": {
            "markdown": r"E_\mathrm{max}",
            "unit": "eV",
            "value": 0.0,
            "lower_bound": 0.0,
            "description": "Energy at maximum TEEY.",
        },
        "delta_E_transition": {
            "markdown": r"\Delta E_{tr}",
            "unit": "eV",
            "value": 1.0,
            "description": "Energy over which we switch from"
            + r" :math:`\sigma_\mathrm{low}` to actual Vaughan TEEY. Useful for"
            + " smoothing the transition. Currently not implemented.",
            "is_locked": True,
        },
        "teey_low": {
            "markdown": r"\sigma_\mathrm{low}",
            "unit": "1",
            "value": 0.5,
            "lower_bound": 0.0,
            "description": "TEEY below :math:`E_0`.",
            "is_locked": True,
        },
        "teey_max": {
            "markdown": r"\sigma_\mathrm{max}",
            "unit": "1",
            "value": 0.0,
            "lower_bound": 0.0,
            "description": "Maximum TEEY, directly taken from the measurement.",
        },
        "k_s": {
            "markdown": r"k_s",
            "unit": "1",
            "value": 1.0,
            "lower_bound": 0.0,
            "upper_bound": 2.0,
            "description": r"Roughness factor (:math:`\sigma_\mathrm{max}`). "
            + " Locked by default, but could be used for more precise fits.",
            "is_locked": True,
        },
        "k_se": {
            "markdown": r"k_{se}",
            "unit": "1",
            "value": 1.0,
            "lower_bound": 0.0,
            "upper_bound": 2.0,
            "description": r"Roughness factor (:math:`E_\mathrm{max}`). "
            + " Locked by default, but could be used for more precise fits.",
            "is_locked": True,
        },
        "E_c1": {
            "markdown": r"E_{c,\,1}",
            "unit": "eV",
            "value": 0.0,
            "description": r"First crossover energy. Must be provided instead"
            + " of E_0 for SPARK3D Vaughan.",
            "is_locked": False,
        },
    }

    def __init__(
        self,
        implementation: VaughanImplementation = "original",
        parameters_values: dict[str, Any] | None = None,
    ) -> None:
        """Instantiate the object.

        .. note::
            Parameter values set by ``implementation`` have priority over
            values given in ``parameters_values``.

        Parameters
        ----------
        implementation:
            Modifies certain presets to match different interpretations of the
            model by calling :meth:`.preset_implementation`. These parameter
            modifications have precedence over the ones set in
            `parameters_values`.
        parameters_values :
            Contains name of parameters and associated value. If provided, will
            override the default values set in ``initial_parameters``.

        """
        super().__init__(url_doc_override="manual/models/vaughan")
        self.parameters: VaughanParameters = {  # type: ignore
            name: Parameter(**kwargs)  # type: ignore
            for name, kwargs in self.initial_parameters.items()
        }
        self._generate_parameter_docs()
        if parameters_values is not None:
            self.set_parameters_values(parameters_values)
        self._func = _vaughan_func
        self.preset_implementation(implementation)

    def preset_implementation(
        self,
        implementation: VaughanImplementation,
    ) -> None:
        r"""Update some parameters to reproduce a specific implementation.

        Vaughan CST:

            - :math:`\sigma_\mathrm{low}` is set to 0.

        Vaughan SPARK3D:

            - :math:`\sigma_\mathrm{low}` is set to 0.
            - :math:`E_0` is unlocked, so that it will be fitted to match
              :math:`E_{c,\,1}`.
            - Below :math:`10` :unit:`eV`, TEEY is 0.

        """
        if implementation == "original":
            return
        if implementation == "CST":
            self.set_parameter_value("teey_low", 0.0)
            return
        if implementation == "SPARK3D":
            self.set_parameters_values(
                {"teey_low": 0.0, "delta_E_transition": 2.0}
            )
            self.parameters["E_0"].is_locked = False

            E_0 = self._E_0_matching(E_c1=self.parameters["E_c1"].value)
            if np.isnan(E_0):
                return
            self.set_parameter_value("E_0", E_0)
            self._func = _vaughan_spark3d

            return
        logging.error(f"{implementation = } not in {VaughanImplementation}")

    def get_data(
        self,
        population: ImplementedPop,
        emission_data_type: ImplementedEmissionData,
        energy: NDArray[np.float64],
        theta: NDArray[np.float64],
        *args,
        **kwargs,
    ) -> pd.DataFrame | None:
        """Return desired data according to current model.

        Will return a dataframe only if the TEEY is asked.

        .. todo::
            This method could be so much simpler and efficient.

        """
        if population != "all" or emission_data_type != "Emission Yield":
            return super().get_data(
                population=population,
                emission_data_type=emission_data_type,
                energy=energy,
                theta=theta,
                *args,
                **kwargs,
            )
        out = np.zeros((len(energy), len(theta)))
        for i, ene in enumerate(energy):
            for j, the in enumerate(theta):
                out[i, j] = self._func(ene, the, **self.parameters)

        out_dict = {f"{the} [deg]": out[:, j] for j, the in enumerate(theta)}
        out_dict["Energy [eV]"] = energy
        return pd.DataFrame(out_dict)

    def find_optimal_parameters(
        self, data_matrix: DataMatrix, **kwargs
    ) -> None:
        """Match with position of first crossover and maximum."""
        if not data_matrix.has_all_mandatory_files(self.model_config):
            logging.info(
                "Files are not all provided. If Ec1 was given, I will try to "
                "find the corresponding E_0."
            )
            self.find_e_0()
            return

        emission_yield = data_matrix.teey
        assert emission_yield.population == "all"

        self.set_parameters_values(
            {
                "E_max": emission_yield.e_max,
                "teey_max": emission_yield.ey_max,
            }
        )
        if not self.parameters["E_c1"].is_locked:
            self.set_parameter_value("E_c1", emission_yield.e_c1)
        if not self.parameters["E_0"].is_locked:
            E_0 = self._E_0_matching(E_c1=self.parameters["E_c1"].value)
            self.set_parameter_value("E_0", E_0)

    def find_e_0(self) -> None:
        """Find E_0 with error handling."""
        e_0 = self.parameters["E_0"]
        assert not e_0.is_locked, "Unlock E_0 to allow for a fit."

        for key in ("E_c1", "E_max", "teey_max"):
            value = self.parameters[key].value
            assert value is not None, f"You must provide a value for {key}"

        E_0 = self._E_0_matching(E_c1=self.parameters["E_c1"].value)
        self.set_parameter_value("E_0", E_0)
        return

    def _E_0_matching(self, *, E_c1: float) -> float:
        """Fit E_0 to retrieve E_c1 (SPARK3D)"""
        parameters = self.parameters.copy()

        def _to_minimize(E_0: float) -> float:
            parameters["E_0"].value = E_0
            teey_at_crossover = _vaughan_func(ene=E_c1, the=0.0, **parameters)
            if isinstance(teey_at_crossover, np.ndarray):
                teey_at_crossover = teey_at_crossover[0]
            return abs(teey_at_crossover - 1.0)

        optimized_E_0 = least_squares(_to_minimize, x0=12.5).x
        return float(optimized_E_0[0])

    def evaluate(self, data_matrix: DataMatrix) -> dict[str, float]:
        """Evaluate the quality of the model using Fil criterions.

        Fil criterions :cite:`Fil2016a,Fil2020` are adapted to TEEY models.

        """
        return self._evaluate_for_teey_models(data_matrix)


def _vaughan_func(
    ene: float,
    the: float,
    E_0: Parameter,
    E_max: Parameter,
    teey_max: Parameter,
    teey_low: Parameter,
    k_se: Parameter,
    k_s: Parameter,
    delta_E_transition: Parameter,
    **parameters,
) -> float | NDArray[np.float64]:
    """Compute the TEEY for incident energy E."""
    mod_e_max = E_max.value * (
        1.0 + k_se.value * math.radians(the) ** 2 / (2.0 * math.pi)
    )
    mod_teey_max = teey_max.value * (
        1.0 + k_s.value * math.radians(the) ** 2 / (2.0 * math.pi)
    )
    if ene < E_0.value:
        return teey_low.value

    xi = (ene - E_0.value) / (mod_e_max - E_0.value)

    if xi <= 1.0:
        k = 0.56
    elif xi <= 3.6:
        k = 0.25
    else:
        return mod_teey_max * 1.125 / (xi**0.35)

    return mod_teey_max * (xi * np.exp(1.0 - xi)) ** k


def _vaughan_spark3d(
    ene: float,
    the: float,
    E_0: Parameter,
    E_max: Parameter,
    teey_max: Parameter,
    teey_low: Parameter,
    k_se: Parameter,
    k_s: Parameter,
    delta_E_transition: Parameter,
    **parameters,
) -> float | NDArray[np.float64]:
    r"""Compute TEEY as SPARK3D would.

    This is a classic Vaughan, but TEEY is null for energies below
    ``E_0_SPARK3D=10.0``. This parameter is different from the classic ``E_0``
    that appears in the expression of :math:`\xi`.

    """
    if ene >= E_0_SPARK3D:
        return _vaughan_func(
            ene=ene,
            the=the,
            E_0=E_0,
            E_max=E_max,
            teey_max=teey_max,
            teey_low=teey_low,
            k_se=k_se,
            k_s=k_s,
            delta_E_transition=delta_E_transition,
            **parameters,
        )
    return teey_low.value


# Append dynamically generated docs to the module docstring
if __doc__ is None:
    __doc__ = ""
__doc__ += Vaughan._generate_parameter_docs()



"""Define an object to store an emission energy distribution."""

from pathlib import Path
from typing import Self

import pandas as pd
from eemilib.emission_data.emission_data import EmissionData
from eemilib.loader.loader import Loader
from eemilib.plotter.plotter import Plotter
from eemilib.util.constants import (
    ImplementedPop,
    col_energy,
    col_normal,
    md_energy_distrib,
)


class EmissionEnergyDistribution(EmissionData):
    """An emission energy distribution."""

    def __init__(
        self,
        population: ImplementedPop,
        data: pd.DataFrame,
        e_pe: float | None = None,
        norm: float | None = None,
    ) -> None:
        """Instantiate the data.

        Parameters
        ----------
        population :
            The concerned population of electrons.
        data :
            Structure holding the data. Must have a ``Energy (eV)`` column
            holding ``population`` energy. And one or several columns
            ``theta [deg]``, where ``theta`` is the value of the incidence
            angle and content is corresponding emission energy.
        e_pe :
            Energy of primary electrons in :unit:`eV`.
        norm :
            To specify re-normalization constant. If not provided, we try to
            set the maximum of SEs to unity. Provide ``1.0`` to avoid any
            normalization.

        """
        super().__init__(population, data)
        self.energies = data[col_energy].to_numpy()
        self.angles = [
            float(col.split()[0]) for col in data.columns if col != col_energy
        ]

        #: Energy at the maximum of SEs in :unit:`eV`. Defined for SEs and
        #: distribution of all electrons.
        self.e_peak_se: float
        i_peak_se, self.e_peak_se = self._find_SE_peak()
        #: Energy at the maximum of EBEs in :unit:`eV`. Defined for EBEs and
        #: distribution of all electrons.
        self.e_peak_ebe: float
        #: Position of EBE peak.
        self.i_peak_ebe: int
        self.i_peak_ebe, self.e_peak_ebe = self._find_EBE_peak()

        #: Energy of PEs in :unit:`eV`. If this information is not found in
        #: the file header, we set it to the value of ``self.e_peak_ebe``.
        self.e_pe: float
        if e_pe:
            self.e_pe = e_pe
        self.e_pe = e_pe if e_pe else self.e_peak_ebe

        #: Re-normalization factor of distribution.
        self.norm: float = (
            norm if norm else self.data.at[i_peak_se, col_normal]
        )
        self._normalize()

    @classmethod
    def from_filepath(
        cls,
        population: ImplementedPop,
        loader: Loader,
        *filepath: str | Path,
    ) -> Self:
        """Instantiate the data from files.

        Parameters
        ----------
        loader :
            The object that will load the data.
        population :
            The concerned population of electrons.
        *filepath :
            Path(s) to file holding data under study.

        """
        data, e_pe = loader.load_emission_energy_distribution(*filepath)
        return cls(population, data, e_pe=e_pe)

    @property
    def label(self) -> str:
        """Print nature of data (markdown)."""
        return md_energy_distrib[self.population]

    def plot[T](
        self,
        plotter: Plotter,
        *args,
        lw: float | None = 0.0,
        marker: str | None = "+",
        axes: T | None = None,
        grid: bool = True,
        population: ImplementedPop | None = None,
        **kwargs,
    ) -> T:
        """Plot the contained data using plotter.

        This wrapper simply calls the
        :meth:`.Plotter.plot_emission_energy_distribution` method.
        method.

        """
        return plotter.plot_emission_energy_distribution(
            df=self.data,
            *args,
            axes=axes,
            lw=lw,
            marker=marker,
            grid=grid,
            label=self.label,
            population=population,
            **kwargs,
        )

    def _normalize(self) -> None:
        """Normalize the distribution."""
        data_columns = [c for c in self.data.columns if c != col_energy]
        self.data[data_columns] /= self.norm

    @property
    def _se_ebe_limit(self) -> int:
        """Arbitrary index limit between SEs and EBEs."""
        return int(self._n_points / 4)

    def _find_SE_peak(self) -> tuple[int, float]:
        """Find the SEs maximum."""
        i = self.data[: self._se_ebe_limit][col_normal].argmax()
        e_peak_se = self.data.at[i, col_energy]
        return int(i), float(e_peak_se)

    def _find_EBE_peak(self) -> tuple[int, float]:
        """Find the position of the EBE peak."""
        i = (
            self.data[self._se_ebe_limit :][col_normal].argmax()
            + self._se_ebe_limit
        )
        e_peak_ebe = self.data.at[i, col_energy]
        return int(i), float(e_peak_ebe)



"""Define an object to store an emission angle distribution."""

from pathlib import Path
from typing import Self

import pandas as pd
from eemilib.emission_data.emission_data import EmissionData
from eemilib.loader.loader import Loader
from eemilib.util.constants import ImplementedPop


class EmissionAngleDistribution(EmissionData):
    """An emission angle distribution."""

    def __init__(
        self,
        population: ImplementedPop,
        data: pd.DataFrame,
    ) -> None:
        """Instantiate the data.

        Parameters
        ----------
        population :
            The concerned population of electrons.
        data :
            Structure holding the data. Must have an ``Angle [deg]`` column
            holding PEs angle. And one or several columns ``theta [deg]``,
            where `theta` is the value of the incidence angle and content is
            corresponding emission angle.

        """
        super().__init__(population, data)
        raise NotImplementedError

    @classmethod
    def from_filepath(
        cls,
        population: ImplementedPop,
        loader: Loader,
        *filepath: str | Path,
    ) -> Self:
        """Instantiate the data from files.

        Parameters
        ----------
        loader :
            The object that will load the data.
        population :
            The concerned population of electrons.
        *filepath :
            Path(s) to file holding data under study.

        """
        data = loader.load_emission_angle_distribution(*filepath)
        return cls(population, data)

    @property
    def label(self) -> str:
        """Print nature of data (markdown)."""
        raise NotImplementedError



"""Define the ABC for the :class:`Loader`."""

from abc import ABC, abstractmethod
from collections.abc import Collection
from pathlib import Path

import pandas as pd
from eemilib.util.helper import documentation_url

paths = Path | str


class Loader(ABC):
    """Define the base class for loading various electron emission files."""

    def __init__(self) -> None:
        """Instantiate the object."""
        self.doc_url = documentation_url(self)

    @abstractmethod
    def load_emission_yield(
        self,
        filepath: str | Path | Collection[str] | Collection[Path],
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Load the given electron emission yield file."""

    @abstractmethod
    def load_emission_energy_distribution(
        self,
        filepath: str | Path | Collection[str] | Collection[Path],
        *args,
        **kwargs,
    ) -> tuple[pd.DataFrame, float | None]:
        """Load the given electron emission energy distribution file.

        Should also try to load the energy of primary electrons from file
        metadata. If not possible, :class:`.EmissionEnergyDistribution` will
        try to infer it from the position of the EBEs peak.

        """

    @abstractmethod
    def load_emission_angle_distribution(
        self,
        filepath: str | Path | Collection[str] | Collection[Path],
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Load the given electron emission angle distribution file."""



"""Define a generic worflow."""

from pathlib import Path

import numpy as np
from eemilib.emission_data.data_matrix import DataMatrix
from eemilib.loader.pandas_loader import PandasLoader
from eemilib.model.vaughan import Vaughan
from eemilib.plotter.pandas import PandasPlotter


def main() -> None:
    """Define the simple basic workflow to adapt as GUI."""
    filepath = Path("../../data/example_copper/measured_TEEY_Cu_1_eroded.txt")
    filepaths = (filepath,)

    # =========================================================================
    # Horizontal screen portion 1 of 3
    # =========================================================================
    data_matrix = DataMatrix()

    # A matrix with 4 rows, 3 cols. You click on a cell, a window opens and you
    # select the file(s) to open.
    data_matrix.set_files(
        filepaths,
        row=3,
        col=0,
    )

    # Dropdown menu. Possible values are modules in eemilib.loader, deriving
    # from Loader
    loader = PandasLoader()

    # "Load" button
    data_matrix.load_data(loader)

    # =========================================================================
    # Horizontal screen portion 2 of 3
    # =========================================================================
    # A dropdown menu. Possible values are modules in eemilib.model, deriving
    # from Model
    model = Vaughan()

    # A matrix with 6 columns, n rows
    # There is one row per (key, value) pair in Model.parameters attribute
    # The list of columns is:
    # 0: parameter.markdown
    # 1: parameter.unit
    # 2: parameter.value (editable, editing it updates parameter.value parameter)
    # 3: parameter.lower_bound (editable, editing it updates parameter.lower_bound parameter)
    # 4: parameter.upper_bound (editable, editing it updates parameter.upper_bound parameter)
    # 5: a "lock" checkbox. checking it calls the parameter.lock method. unchecking it calls the paramer.unlock method

    # The "Fit!" button, at the bottom
    model.find_optimal_parameters(data_matrix)
    # It will update the values of model.parameters[key].value, so the matrix
    # above should be updated too

    # =========================================================================
    # Horizontal screen, portion 3 of 3
    # =========================================================================
    # A dropdown menu. Possible values are modules in eemilib.plotter, deriving
    # from Plotter
    plotter = PandasPlotter()

    # Cases to tick (several values possible); the possible values are in the
    # tuple constants.ImplementedPop
    population = "all"
    # Case to tick (one value possible); the possible values are in
    # constants.ImplementedEmissionData
    emission_data_type = "Emission Yield"

    # The "Plot measured" button
    axes = data_matrix.plot(
        plotter, population=population, emission_data_type=emission_data_type
    )

    # Energy [eV]: (here start, stop, nstep boxes)
    e_start = 0
    e_end = 1000
    n_e = 1001

    # Angles [deg]: (here start, stop, nstep boxes)
    theta_start = 0
    theta_end = 60
    n_theta = 4

    # The "Plot model" button
    axes = model.plot(
        plotter,
        population=population,
        emission_data_type=emission_data_type,
        energies=np.linspace(e_start, e_end, n_e),
        angles=np.linspace(theta_start, theta_end, n_theta),
        axes=axes,
    )

    # Evaluate
    evaluations = model.evaluate(data_matrix)
    print(evaluations)
    return


if __name__ == "__main__":
    main()



"""Python dual-logging setup (console and log file).

It supports different log levels and colorized output.

Created by Fonic <https://github.com/fonic>
Date: 04/05/20 - 02/07/23

Based on:
https://stackoverflow.com/a/13733863/1976617
https://uran198.github.io/en/python/2016/07/12/colorful-python-logging.html
https://en.wikipedia.org/wiki/ANSI_escape_code#Colors

"""

import logging
import subprocess
import sys
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from typing import Literal


def _get_package_version(package_name: str) -> str:
    """Get the package version."""
    try:
        return version(package_name)
    except PackageNotFoundError:
        return "Unknown version"


def _get_last_commit_hash() -> str:
    """Get the last Git commit hash.

    .. todo::
        Will look for the script commit number, not for the code commit number!

    """
    try:
        return (
            subprocess.check_output(["git", "rev-parse", "HEAD"])
            .decode("utf-8")
            .strip()
        )
    except subprocess.CalledProcessError:
        return "Unknown commit"


def _log_header(package_name: str) -> str:
    """Create a header for the log file."""
    package_version = _get_package_version(package_name)
    commit_hash = _get_last_commit_hash()
    header_message = (
        f"Starting log for {package_name} - Version: {package_version}, "
        f"Commit: {commit_hash}"
    )
    return header_message


def _console_handler(
    output: str, level: str, color: bool, line_template: str
) -> logging.Handler:
    """Set up the console handler."""
    output_stream = sys.stdout if output.lower() == "stdout" else sys.stderr
    console_handler = logging.StreamHandler(output_stream)
    console_handler.setLevel(level.upper())
    console_formatter = LogFormatter(fmt=line_template, color=color)
    console_handler.setFormatter(console_formatter)
    return console_handler


def _file_handler(
    file: Path, level: str, color: bool, line_template: str
) -> logging.Handler | Literal[False]:
    """Set up the file handler."""
    try:
        logfile_handler = logging.FileHandler(file, mode="a")
    except Exception as e:
        print(f"Failed to set up log file: {e}")
        return False

    logfile_handler.setLevel(level.upper())
    logfile_formatter = LogFormatter(fmt=line_template, color=color)
    logfile_handler.setFormatter(logfile_formatter)
    return logfile_handler


class LogFormatter(logging.Formatter):
    """Logging formatter supporting colorized output."""

    COLOR_CODES = {
        # bright/bold magenta
        logging.CRITICAL: "\033[1;35m",
        # bright/bold red
        logging.ERROR: "\033[1;31m",
        # bright/bold yellow
        logging.WARNING: "\033[1;33m",
        # white / light gray
        logging.INFO: "\033[0;37m",
        # bright/bold black / dark gray
        logging.DEBUG: "\033[1;30m",
    }

    RESET_CODE = "\033[0m"

    def __init__(self, color: bool, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.color = color

    def format(self, record: logging.LogRecord, *args, **kwargs) -> str:
        if self.color and record.levelno in self.COLOR_CODES:
            record.color_on = self.COLOR_CODES[record.levelno]
            record.color_off = self.RESET_CODE
        else:
            record.color_on = ""
            record.color_off = ""
        return super().format(record, *args, **kwargs)


def set_up_logging(
    package_name: str,
    console_log_output: str = "stdout",
    console_log_level: str = "INFO",
    console_log_color: bool = True,
    console_log_line_template: str = "%(color_on)s[%(levelname)-8s] [%(filename)-20s]%(color_off)s %(message)s",
    logfile_file: Path = Path("eemilib.log"),
    logfile_log_level: str = "INFO",
    logfile_log_color: bool = False,
    logfile_line_template: str = "%(color_on)s[%(asctime)s] [%(levelname)-8s] [%(filename)-20s]%(color_off)s %(message)s",
) -> bool:
    """Set up logging with both console and file handlers."""
    # Remove previous logger
    for handler in logging.root.handlers[:]:
        handler.close()
        logging.root.removeHandler(handler)

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    logger.addHandler(
        _console_handler(
            console_log_output,
            console_log_level,
            console_log_color,
            console_log_line_template,
        )
    )

    logfile_handler = _file_handler(
        logfile_file,
        logfile_log_level,
        logfile_log_color,
        logfile_line_template,
    )
    if not logfile_handler:
        return False
    logger.addHandler(logfile_handler)
    logger.info(_log_header(package_name))

    return True


def main():
    """Main function."""
    if not set_up_logging(
        package_name="EEmiLib",
        console_log_output="stdout",
        console_log_level="warning",
        console_log_color=True,
        logfile_file=Path("eemilib.log"),
        logfile_log_level="INFO",
        logfile_log_color=False,
    ):
        print("Failed to set up logging, aborting.")
        return 1

    # Sample log messages
    logging.debug("Debug message")
    logging.info("Info message")
    logging.warning("Warning message")
    logging.error("Error message")
    logging.critical("Critical message")
    return 0



"""Create the Chung and Everhart model, to compute SEs emission distribution.

You will need to provice emission energy distribution measurements.

"""

from typing import Any, TypedDict

import numpy as np
import pandas as pd
from eemilib.core.model_config import ModelConfig
from eemilib.emission_data.data_matrix import DataMatrix
from eemilib.model.model import Model
from eemilib.model.parameter import Parameter
from eemilib.util.constants import (
    ImplementedEmissionData,
    ImplementedPop,
    col_energy,
    col_normal,
)
from numpy.typing import NDArray
from scipy.optimize import least_squares


class ChungEverhartParameters(TypedDict):
    W_f: Parameter
    norm: Parameter


class ChungEverhart(Model):
    """Define the Chung and Everhart model, defined in :cite:`Chung1974`."""

    emission_data_types = ["Emission Energy"]
    populations = ["SE"]
    considers_energy = True
    is_3d = False
    is_dielectrics_compatible = False
    model_config = ModelConfig(
        emission_yield_files=(),
        emission_energy_files=("all",),
        emission_angle_files=(),
    )
    initial_parameters = {
        "W_f": {
            "markdown": r"W_f",
            "unit": "eV",
            "value": 8.0,
            "lower_bound": 0.0,
            "description": "Material work function.",
        },
        "norm": {
            "markdown": r"k",
            "unit": "1",
            "value": 1.0,
            "lower_bound": 0.0,
            "description": "Distribution re-normalization constant.",
        },
    }

    def __init__(
        self, parameters_values: dict[str, Any] | None = None
    ) -> None:
        """Instantiate the object.

        Parameters
        ----------
        parameters_values :
            Contains name of parameters and associated value. If provided, will
            override the default values set in ``initial`_parameters``.

        """
        super().__init__(url_doc_override="manual/models/chung_and_everhart")
        self.parameters: ChungEverhartParameters = {  # type: ignore
            name: Parameter(**kwargs)  # type: ignore
            for name, kwargs in self.initial_parameters.items()
        }
        self._generate_parameter_docs()
        if parameters_values is not None:
            self.set_parameters_values(parameters_values)

        self._func = _chung_everhart_func

    def get_data(
        self,
        population: ImplementedPop,
        emission_data_type: ImplementedEmissionData,
        energy: NDArray[np.float64],
        theta: NDArray[np.float64],
        *args,
        **kwargs,
    ) -> pd.DataFrame | None:
        """Return desired data according to current model.

        Will return a dataframe only if the SEs energy distribution is asked.

        """
        if population != "SE" or emission_data_type != "Emission Energy":
            return super().get_data(
                population=population,
                emission_data_type=emission_data_type,
                energy=energy,
                theta=theta,
                *args,
                **kwargs,
            )
        out = np.zeros(len(energy))
        for i, ene in enumerate(energy):
            out[i] = self._func(
                ene, W_f=self.parameters["W_f"], norm=self.parameters["norm"]
            )

        out_dict = {col_normal: out, col_energy: energy}
        return pd.DataFrame(out_dict)

    def find_optimal_parameters(
        self, data_matrix: DataMatrix, **kwargs
    ) -> None:
        """Fit model parameters on measurements."""
        if not data_matrix.has_all_mandatory_files(self.model_config):
            raise ValueError("Files are not all provided.")

        distribution = data_matrix.all_energy_distribution
        assert distribution.population == "all"

        lsq = least_squares(
            fun=_residue,
            x0=8.0,
            bounds=(0, np.inf),
            args=(
                distribution.data[col_energy].to_numpy(),
                distribution.data[col_normal].to_numpy(),
            ),
        )
        w_f = lsq.x[0]
        self.set_parameters_values(
            {"W_f": w_f, "norm": _chung_everhart_norm(w_f)}
        )


def _chung_everhart_norm(w_f: float) -> float:
    """Return norm value to have distribution maximum to unity."""
    return 256.0 * w_f**3 / 27.0


def _chung_everhart_func(
    ene: float | NDArray[np.float64],
    W_f: Parameter | float,
    norm: Parameter | None = None,
    **parameters,
) -> float | NDArray[np.float64]:
    """Compute the energy distribution."""
    w_f_value = W_f.value if isinstance(W_f, Parameter) else W_f
    norm_value = (
        norm.value if norm is not None else _chung_everhart_norm(w_f_value)
    )
    return norm_value * ene / (ene + w_f_value) ** 4


def _residue(
    w_f: float, ene: NDArray[np.float64], measured: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Compute array of residues between model and measurements."""
    return _chung_everhart_func(ene, w_f) - measured


# Append dynamically generated docs to the module docstring
if __doc__ is None:
    __doc__ = ""
__doc__ += ChungEverhart._generate_parameter_docs()



"""Create the Sombrin model, to compute TEEY.

It is very precise on the first crossover energy, but does not take the
incident angle into account.

"""

import math
from typing import Any, TypedDict

import numpy as np
import pandas as pd
from eemilib.core.model_config import ModelConfig
from eemilib.emission_data.data_matrix import DataMatrix
from eemilib.model.model import Model
from eemilib.model.parameter import Parameter
from eemilib.util.constants import (
    ImplementedEmissionData,
    ImplementedPop,
    col_energy,
    col_normal,
)
from numpy.typing import NDArray


class SombrinParameters(TypedDict):
    E_max: Parameter
    teey_max: Parameter
    E_c1: Parameter


class Sombrin(Model):
    """Define the Sombrin model, defined in :cite:`Sombrin1993`.

    We use the implementation from :cite:`Fil2016a`.

    """

    emission_data_types = ["Emission Yield"]
    populations = ["all"]
    considers_energy = True
    is_3d = False
    is_dielectrics_compatible = False
    model_config = ModelConfig(
        emission_yield_files=("all",),
        emission_energy_files=(),
        emission_angle_files=(),
    )
    initial_parameters = {
        "E_max": {
            "markdown": r"E_\mathrm{max}",
            "unit": "eV",
            "value": 1.0,
            "lower_bound": 0.0,
            "description": "Energy at maximum TEEY.",
        },
        "teey_max": {
            "markdown": r"\sigma_\mathrm{max}",
            "unit": "1",
            "value": 0.0,
            "lower_bound": 0.0,
            "description": "Maximum TEEY, directly taken from the measurement.",
        },
        "E_c1": {
            "markdown": r"E_{c,\,1}",
            "unit": "eV",
            "value": 0.0,
            "lower_bound": 0.0,
            "description": (
                "First crossover energy. Must be provided instead of E_0 for "
                "SPARK3D Vaughan."
            ),
        },
    }

    def __init__(
        self, parameters_values: dict[str, Any] | None = None
    ) -> None:
        """Instantiate the object.

        Parameters
        ----------
        parameters_values :
            Contains name of parameters and associated value. If provided, will
            override the default values set in ``initial_parameters``.

        """
        super().__init__(url_doc_override="manual/models/sombrin")
        self.parameters: SombrinParameters = {  # type: ignore
            name: Parameter(**kwargs)  # type: ignore
            for name, kwargs in self.initial_parameters.items()
        }
        self._generate_parameter_docs()
        if parameters_values is not None:
            self.set_parameters_values(parameters_values)

        self._func = _sombrin_func
        self._E: float | None = None

    @property
    def E(self) -> float:
        """Return the ``E`` parameter in Sombrin model. Not incident energy."""
        if self._E is not None:
            return self._E
        self._E = _e_parameter(
            self.parameters["teey_max"],
            self.parameters["E_max"],
            self.parameters["E_c1"],
        )
        return self._E

    def get_data(
        self,
        population: ImplementedPop,
        emission_data_type: ImplementedEmissionData,
        energy: NDArray[np.float64],
        theta: NDArray[np.float64],
        *args,
        **kwargs,
    ) -> pd.DataFrame | None:
        """Return desired data according to current model.

        Will return a dataframe only if the TEEY is asked.

        """
        if population != "all" or emission_data_type != "Emission Yield":
            return super().get_data(
                population=population,
                emission_data_type=emission_data_type,
                energy=energy,
                theta=theta,
                *args,
                **kwargs,
            )
        out = np.zeros(len(energy))
        for i, ene in enumerate(energy):
            out[i] = self._func(
                ene,
                E_max=self.parameters["E_max"],
                teey_max=self.parameters["teey_max"],
                E_c1=self.parameters["E_c1"],
                E_param=self.E,
            )

        out_dict = {col_normal: out, col_energy: energy}
        return pd.DataFrame(out_dict)

    def set_parameter_value(self, name: str, value: Any) -> None:
        """Set ``E`` to None before updating the parameter."""
        if self._E is not None:
            self._E = None
        return super().set_parameter_value(name, value)

    def find_optimal_parameters(
        self, data_matrix: DataMatrix, **kwargs
    ) -> None:
        """Extract main TEEY curve parameters from measure."""
        if not data_matrix.has_all_mandatory_files(self.model_config):
            raise ValueError("Files are not all provided.")

        emission_yield = data_matrix.teey
        assert emission_yield.population == "all"

        self.set_parameters_values(
            {
                "E_max": emission_yield.e_max,
                "teey_max": emission_yield.ey_max,
                "E_c1": emission_yield.e_c1,
            }
        )

    def evaluate(self, data_matrix: DataMatrix) -> dict[str, float]:
        """Evaluate the quality of the model using Fil criterions.

        Fil criterions :cite:`Fil2016a,Fil2020` are adapted to TEEY models.

        """
        return self._evaluate_for_teey_models(data_matrix)


def _sombrin_func(
    ene: float | NDArray[np.float64],
    E_max: Parameter,
    teey_max: Parameter,
    E_c1: Parameter,
    E_param: float | None,
    **parameters,
) -> float | NDArray[np.float64]:
    """Compute the TEEY for incident energy E."""
    if E_param is None:
        E_param = _e_parameter(teey_max, E_max, E_c1)
    num = 2 * teey_max.value * (ene / E_max.value) ** E_param
    denom = 1 + (ene / E_max.value) ** (2 * E_param)
    return num / denom


def _e_parameter(
    sigma_max: Parameter, E_max: Parameter, E_c1: Parameter
) -> float:
    """Compute parameter ``E`` in Sombrin model."""
    E_param = math.log(
        sigma_max.value - math.sqrt(sigma_max.value**2 - 1)
    ) / math.log(E_c1.value / E_max.value)
    return E_param


# Append dynamically generated docs to the module docstring
if __doc__ is None:
    __doc__ = ""
__doc__ += Sombrin._generate_parameter_docs()



"""Test the helper functions for emission data."""

import numpy as np
import pandas as pd
from eemilib.emission_data.helper import resample, trim
from eemilib.util.constants import col_energy, col_normal


class TestTrim:
    """Test that trimming works."""

    normal_ey = pd.DataFrame(
        {
            col_energy: np.linspace(0.0, 100.0, 11),
            col_normal: np.random.rand(11),
        }
    )

    def test_no_trim(self) -> None:
        """Test no trimming."""
        expected = self.normal_ey
        returned = trim(self.normal_ey)
        assert np.array_equal(expected, returned)

    def test_low_trim(self) -> None:
        """Test lower trimming."""
        expected = self.normal_ey.iloc[2:]
        returned = trim(self.normal_ey, min_e=20)
        assert np.array_equal(expected, returned)

    def test_upp_trim(self) -> None:
        """Test upper trimming."""
        expected = self.normal_ey.iloc[:-2]
        returned = trim(self.normal_ey, max_e=80)
        assert np.array_equal(expected, returned)

    def test_both_trim(self) -> None:
        """Test both trimming."""
        expected = self.normal_ey.iloc[3:-3]
        returned = trim(self.normal_ey, min_e=30, max_e=70)
        assert np.array_equal(expected, returned)


class TestResample:
    """Test that resampling emission yield works."""

    def _generate_fake_ey(self, n_points: int) -> pd.DataFrame:
        """Generate a generic emission yield."""
        assert n_points % 2 == 1
        half = int((n_points + 1) / 2)
        fake_ey = pd.DataFrame(
            {
                col_energy: np.linspace(0, 200, n_points),
                col_normal: np.hstack(
                    (
                        np.linspace(0.5, 2.5, half),
                        np.linspace(2.5, 0.5, half)[1:],
                    )
                ),
            }
        )
        return fake_ey

    def test_resample(self) -> None:
        """Test resampling the emission yield."""
        n_orig, n_resample = 11, 15
        expected = self._generate_fake_ey(n_resample)
        original = self._generate_fake_ey(n_orig)
        returned = resample(original, n_resample)
        print(returned - expected)
        assert np.allclose(returned, expected, atol=1e-15)



"""Define the base object that will store emission data.

.. todo::
    Add an ``interpolate`` or ``resample`` method. Would be used to have more
    points, in particular when there is few points.

"""

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Self

import numpy as np
import pandas as pd
from eemilib.loader.loader import Loader
from eemilib.plotter.plotter import Plotter
from eemilib.util.constants import ImplementedPop
from eemilib.util.helper import documentation_url
from numpy.typing import NDArray


class EmissionData(ABC):
    """A yield, energy distribution or angular distribution."""

    def __init__(
        self,
        population: ImplementedPop,
        data: pd.DataFrame,
    ) -> None:
        """Instantiate the data.

        Parameters
        ----------
        population :
            The concerned population of electrons.
        data :
            Structure holding the data. Column headers as well as units must
            follow specications (see subclasses documentation).

        """
        self.doc_url = documentation_url(self)
        self.population = population
        self.data = data
        self._n_points = len(self.data)

        self.energies: NDArray[np.float64] | list[float]
        self.angles: NDArray[np.float64] | list[float]

    @classmethod
    @abstractmethod
    def from_filepath(
        cls,
        population: ImplementedPop,
        loader: Loader,
        *filepath: str | Path,
    ) -> Self:
        """Instantiate the data from files.

        Parameters
        ----------
        loader :
            The object that will load the data.
        population :
            The concerned population of electrons.
        *filepath :
            Path(s) to file holding data under study.

        """

    @property
    @abstractmethod
    def label(self) -> str:
        """Print markdown info."""

    @abstractmethod
    def plot[T](
        self, plotter: Plotter, *args, axes: T | None = None, **kwargs
    ) -> T:
        """Plot the contained data using plotter."""



"""Define a role for easier and more consistent display of units."""

from __future__ import annotations

from docutils import nodes
from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxRole
from sphinx.util.typing import ExtensionMetadata


class UnitRole(SphinxRole):
    """A role to display units in math's mathrm format.

    Note that in order to show units such as Ohm, the omega must be escaped
    twice: :unit:`\\Omega`.

    """

    def run(self) -> tuple[list[nodes.Node], list[nodes.system_message]]:
        text = "".join((r"\mathrm{", self.text, r"}"))
        node = nodes.math(text=text)
        return [node], []


def setup(app: Sphinx) -> ExtensionMetadata:
    """Plug new directives into Sphinx."""
    app.add_role("unit", UnitRole())

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Define the base class for all electron emission models.

.. todo::
    Define all the properties: EBEEY, emission energy distributions, etc.

"""

import logging
import math
from abc import ABC, abstractmethod
from collections.abc import Collection
from pprint import pformat
from typing import Any

import numpy as np
import pandas as pd
from eemilib.core.model_config import ModelConfig
from eemilib.emission_data.data_matrix import DataMatrix
from eemilib.emission_data.emission_yield import EmissionYield
from eemilib.model.parameter import Parameter
from eemilib.plotter.plotter import Plotter
from eemilib.util.constants import (
    ImplementedEmissionData,
    ImplementedPop,
    col_energy,
    col_normal,
)
from eemilib.util.helper import documentation_url
from numpy.typing import NDArray


class Model(ABC):
    """Define the base electron emission model.

    Parameters
    ----------
    emission_data_types :
        Types of modelled data.
    populations :
        Modelled populations.
    considers_energy :
        Tell if the model has a dependency over PEs impact energy.
    is_3d :
        Tell if the model has a dependency over PEs impact angle.
    is_dielectrics_compatible :
        Tell if the model can take the surface-trapped charges into account.
    initial_parameters :
        List the :class:`.Parameter` kwargs.
    model_config :
        List the files that the model needs to know in order to work.

    """

    emission_data_types: list[ImplementedEmissionData]
    populations: list[ImplementedPop]
    considers_energy: bool
    is_3d: bool
    is_dielectrics_compatible: bool
    initial_parameters: dict[str, dict[str, str | float | bool]]

    model_config: ModelConfig

    def __init__(
        self, *args, parameters_values: dict[str, Any] | None = None, **kwargs
    ) -> None:
        """Instantiate the object.

        Parameters
        ----------
        parameters_values :
            Contains name of parameters and associated value. If provided, will
            override the default values set in ``initial_parameters``.

        """
        self.doc_url = documentation_url(self, **kwargs)
        self.parameters: dict[str, Parameter]

    @classmethod
    def _generate_parameter_docs(cls) -> str:
        """Generate documentation for the :class:`.Parameter`."""
        doc_lines = [
            "",
            "Model parameters",
            "================",
            "",
            ".. list-table::",
            "   :widths: 5 10 5 5 65",
            "   :header-rows: 1",
            "",
            "   * - Parameter",
            "     - Name",
            "     - Unit",
            "     - Initial",
            "     - Description",
        ]
        for name, kwargs in cls.initial_parameters.items():
            doc = [
                f"   * - :math:`{kwargs.get('markdown', '')}`",
                f"     - {name}",
                f"     - :unit:`{kwargs.get('unit', '')}`",
                f"     - :math:`{kwargs.get('value', '')}`",
                f"     - {kwargs.get('description', '')}",
            ]
            doc_lines += doc
        return "\n".join(doc_lines)

    def teey(
        self,
        energy: NDArray[np.float64],
        theta: NDArray[np.float64],
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        r"""Compute TEEY :math:`\sigma`."""
        teey = self.get_data(
            "all",
            "Emission Yield",
            energy=energy,
            theta=theta,
            *args,
            **kwargs,
        )
        if teey is not None:
            return teey
        logging.warning("No TEEY data found, returning dummy.")
        return _dummy_df(energy, theta)

    def seey(
        self,
        energy: NDArray[np.float64],
        theta: NDArray[np.float64],
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        r"""Compute SEEY :math:`\delta`."""
        seey = self.get_data(
            "SE", "Emission Yield", energy=energy, theta=theta, *args, **kwargs
        )
        if seey is not None:
            return seey
        logging.warning("No SEEY data found, returning dummy.")
        return _dummy_df(energy, theta)

    def se_energy_distribution(
        self,
        energy: NDArray[np.float64],
        theta: NDArray[np.float64],
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        r"""Compute SEs emission energy distribution."""
        se_distrib = self.get_data(
            "SE",
            "Emission Energy",
            energy=energy,
            theta=theta,
            *args,
            **kwargs,
        )
        if se_distrib is not None:
            return se_distrib
        logging.warning(
            "No SE energy distribution data found, returning dummy."
        )
        return _dummy_df(energy, theta)

    def get_data(
        self,
        population: ImplementedPop,
        emission_data_type: ImplementedEmissionData,
        energy: NDArray[np.float64],
        theta: NDArray[np.float64],
        *args,
        **kwargs,
    ) -> pd.DataFrame | None:
        """Return desired data according to current model.

        You should override this method for each :class:`.Model` subclass.
        When desired data is not found, a ``None`` is returned. If you want a
        dummy dataframe instead, call the specific methods for every quantity:
        :meth:`.Model.teey`, :meth:`.Model.seey`,
        :meth:`.Model.se_energy_distribution`.

        """
        return None

    @abstractmethod
    def find_optimal_parameters(
        self, data_matrix: DataMatrix, **kwargs
    ) -> None:
        """Find the best parameters for the current model."""

    def plot[T](
        self,
        plotter: Plotter,
        population: ImplementedPop | Collection[ImplementedPop],
        emission_data_type: ImplementedEmissionData,
        energies: NDArray[np.float64],
        angles: NDArray[np.float64],
        axes: T | None = None,
        grid: bool = True,
        **kwargs,
    ) -> T | None:
        """Plot desired modelled data using ``plotter``.

        This method uses :meth:`.Model.get_data` to compute the modelled data
        matching ``population`` and ``emission_data_type``. Then it calls the
        :meth:`.Model.plot` method.

        Parameters
        ----------
        plotter :
            Object realizing the plot. We transfer it to the
            :meth:`.Model.plot` method.
        population :
            One or several populations to plot. If several are given, we simply
            recursively call this method.
        emission_data_type :
            Type of data to plot.
        energies :
            Energies in :unit:`eV` for which model should be plotted.
        angles :
            Angles in :unit:`deg` for which model should be plotted.
        axes :
            Axes to re-use if given.
        grid :
            If grid should be plotted.
        kwargs :
            Other keyword arguments passed to the :meth:`.Model.plot`
            method.

        Returns
        -------
            Created axes object, or ``None`` if no plot was created.

        """
        if isinstance(population, Collection) and not isinstance(
            population, str
        ):
            for pop in population:
                axes = self.plot(
                    plotter,
                    pop,
                    emission_data_type,
                    energies,
                    angles,
                    axes=axes,
                    grid=grid,
                    **kwargs,
                )
            return axes

        to_plot = self.get_data(
            population=population,
            emission_data_type=emission_data_type,
            energy=energies,
            theta=angles,
        )
        if to_plot is None:
            logging.info(
                f"No modelled data found for {population = } and "
                f"{emission_data_type = }. Skipping this plot."
            )
            return axes

        if emission_data_type == "Emission Yield":
            return plotter.plot_emission_yield(
                to_plot,
                axes=axes,
                ls="--",
                grid=grid,
                population=population,
                **kwargs,
            )
        if emission_data_type == "Emission Energy":
            return plotter.plot_emission_energy_distribution(
                to_plot,
                axes=axes,
                ls="--",
                grid=grid,
                population=population,
                **kwargs,
            )
        raise NotImplementedError

    def set_parameter_value(self, name: str, value: Any) -> None:
        """Give the parameter named ``name`` the value ``value``."""
        if name not in self.parameters:
            logging.warning(
                f"{name = } is not defined for {self}. Skipping... "
            )
            return
        self.parameters[name].value = value

    def set_parameters_values(self, values: dict[str, Any]) -> None:
        """Set multiple parameter values."""
        for name, value in values.items():
            self.set_parameter_value(name, value)

    def evaluate(
        self, data_matrix: DataMatrix, *args, **kwargs
    ) -> dict[str, float]:
        """Evaluate the precision of the model w.r.t. given data."""
        raise NotImplementedError

    def _evaluate_for_teey_models(
        self, data_matrix: DataMatrix
    ) -> dict[str, float]:
        """Evaluate a TEEY model with N. Fil criterions.

        Ref: :cite:`Fil2016a,Fil2020`

        """
        emission_yield = data_matrix.teey
        errors = {
            r"Relative error over $E_{c1}$ [%]": self._error_ec1(
                emission_yield
            ),
            r"$\sigma$ deviation between $E_{c1}$ and $E_{max}$ [%]": self._error_teey(
                emission_yield
            ),
        }
        return errors

    def _error_ec1(self, emission_yield: EmissionYield) -> float:
        """Compute relative error over first crossover energy in :unit:`%`."""
        measured_ec1 = emission_yield.e_c1
        energy = np.linspace(0, 1.5 * measured_ec1, 10001, dtype=np.float64)
        theta = np.array([0.0])
        teey = self.teey(energy, theta)

        idx_ec1 = (teey[col_normal] - 1.0).abs().idxmin()
        model_ec1 = energy[idx_ec1]

        std = math.sqrt((measured_ec1 - model_ec1) ** 2)
        error = 100.0 * std / measured_ec1
        return float(error)

    def _error_teey(self, emission_yield: EmissionYield) -> float:
        """Compute TEEY relative error between E_c1 and E_max in :unit:`%`."""
        min_energy = emission_yield.e_c1
        max_energy = emission_yield.e_max
        df = emission_yield.data
        mask = (df[col_energy] >= min_energy) & (df[col_energy] <= max_energy)

        measured_teey = df.loc[mask, col_normal].to_numpy()
        measured_energy = df.loc[mask, col_energy].to_numpy()
        angles = np.array([0.0])
        modelled_teey = self.teey(measured_energy, angles)[
            col_normal
        ].to_numpy()

        error = 100.0 * np.std(measured_teey - modelled_teey, ddof=1.0)
        return float(error)

    def display_parameters(self) -> None:
        """Display the parameters and their values in a nice looking way."""
        if not hasattr(self, "parameters"):
            logging.info("`parameters` attribute was not set.")

        msg = {
            f"{key:>20}": str(param) for key, param in self.parameters.items()
        }
        logging.info("Parameters values:\n" + pformat(msg))


def _dummy_df(
    energy: NDArray[np.float64], theta: NDArray[np.float64]
) -> pd.DataFrame:
    """Return a null array with proper shape."""
    n_energy = len(energy)
    n_theta = len(theta)
    out = np.zeros((n_energy, n_theta))
    out_dict = {f"{the} [deg]": out[:, j] for the, j in enumerate(theta)}
    out_dict["Energy [eV]"] = energy
    return pd.DataFrame(out_dict)



"""This set of data is available online :cite:`Placais2020b`. Please use this
reference if you use it in your work.

It corresponds to the two Cu samples I used in my PhD :cite:`Placais2021`.

This data can be loaded with :class:`.pandas_loader.PandasLoader`.

"""



"""Define some helper functions."""

import logging
from collections.abc import Sequence

from eemilib.util.constants import (
    ImplementedEmissionData,
    ImplementedPop,
    col_energy,
    md_energy_distrib,
    md_ey,
)


def explicit_column_names(
    columns: Sequence[str],
    population: ImplementedPop | None = None,
    emission_data_type: ImplementedEmissionData | None = None,
) -> dict[str, str]:
    """Explicit column names for the plot.

    This is used to have clearer legends in the plot.

    Parameters
    ----------
    columns :
        Columns of the data frame to be plotted.
    population :
        Type of emitted electrons in data frame.
    emission_data_type :
        Type of data stored in data frame.

    Returns
    -------
    dict[str, str]
        Mapping to easily rename the data frame.

    """
    if population is None:
        logging.info(
            "Cannot explicit column names as population kwargs was not given."
            " Keeping original."
        )
        return {col: col for col in columns}

    if emission_data_type == "Emission Yield":
        explicit = {
            col: (
                f"{md_ey[population]} @{col}"
                if col != col_energy
                else "PEs energy [eV]"
            )
            for col in columns
        }
        return explicit

    if emission_data_type == "Emission Energy":
        explicit = {
            col: (
                f"{md_energy_distrib[population]} @{col}"
                if col != col_energy
                else col_energy
            )
            for col in columns
        }
        return explicit

    logging.info(
        f"Explicit column names not implemented for {emission_data_type = }. "
        "Keeping original."
    )
    return {col: col for col in columns}



"""Define a loader adapted to DEESSE (ONERA, Toulouse) file format."""

import logging
from pathlib import Path
from typing import Any

import pandas as pd
from eemilib.loader.loader import Loader
from eemilib.util.constants import col_energy, col_normal


class DeesseLoader(Loader):
    """Define the loader."""

    def __init__(self) -> None:
        """Raise an error for now.

        Ideally, this loader should detect correct input and columns. But it is
        not for now.

        """
        super().__init__()

    def load_emission_yield(self, *filepath: str | Path) -> pd.DataFrame:
        """Load and format the given emission yield files.

        Parameters
        ----------
        filepath :
            Path(s) to file holding data under study.

        Returns
        -------
        pandas.DataFrame
            Structure holding the data. Has a ``Energy [eV]`` column
            holding PEs energy. And one or several columns ``theta [deg]``,
            where `theta` is the value of the incidence angle and content is
            corresponding emission yield.

        """
        col1 = "Energie réelle des électrons (eV)"
        col2 = "TEEY"
        kwargs = {
            "sep": ";",
            "encoding": "latin1",
            "header": 5,
        }
        all_df = []
        for file in filepath:
            full_df = pd.read_csv(file, **kwargs)
            incidence_angle = self._extract_incidence_angle(full_df)
            of_interest_df = full_df[[col1, col2]].rename(
                columns={col1: col_energy, col2: f"{incidence_angle} [deg]"}
            )
            all_df.append(of_interest_df.set_index(col_energy))

        concatenated = pd.concat(all_df, axis=1)
        logging.info(f"Successfully loaded emission yield file(s) {filepath}")
        return concatenated.reset_index()

    def _extract_incidence_angle(self, full_data: pd.DataFrame) -> float:
        """Try to get the incidence angle in the file."""
        row_number = 5
        col_number = -1
        angle_as_str = full_data.iloc[row_number].iloc[col_number]
        assert isinstance(angle_as_str, str)
        try:
            angle = float(angle_as_str)
        except ValueError:
            angle = float(angle_as_str[:-1])
        return angle

    def load_emission_angle_distribution(self, *args) -> Any:
        raise NotImplementedError

    def load_emission_energy_distribution(
        self,
        filepath: str | Path,
        e_pe: float | None = None,
    ) -> tuple[pd.DataFrame, float | None]:
        """Load and format an emission energy file from DEESSE.

        Parameters
        ----------
        filepath :
            Path to file holding data under study.
        e_pe :
            Energy of PEs in :unit:`eV`. Should be manually provided, as not
            present in DEESSE files.

        Returns
        -------
        pandas.DataFrame
            Structure holding the data. Has a ``Energy [eV]`` column
            holding emitted electrons energy. And one or several columns
            ``theta [deg]``, where ``theta`` is the value of the incidence
            angle and content is corresponding emission energy distribution.
        float
            Energy of Primary Electrons in :unit:`eV`. If not found in the file
            comments, it will be inferred from the position of the EBEs peak.

        """
        col1 = "Kinetic Energy [eV]"
        col2 = "Intensity[cts/s]"
        extension = Path(filepath).suffix
        if extension == ".csv":
            df = pd.read_csv(filepath, sep=";")
        elif extension == ".xlsx":
            df = pd.read_excel(filepath)
        else:
            raise RuntimeError(f"Filetype of {filepath} is not supported.")
        df = df[[col1, col2]].rename(
            columns={col1: col_energy, col2: col_normal}
        )
        return df, e_pe



"""Define plotter relying on pandas."""

import matplotlib.pyplot as plt
import pandas as pd
from eemilib.plotter.helper import explicit_column_names
from eemilib.plotter.plotter import Plotter
from eemilib.util.constants import ImplementedPop, col_energy, md_ylabel
from matplotlib.axes import Axes


class PandasPlotter(Plotter):
    """A :class:`.Plotter` using pandas lib."""

    def __init__(self, *args, gui: bool = False, **kwargs) -> None:
        """Instantiate object.

        Parameters
        ----------
        gui :
            Activates interactive plotting if using GUI.

        """
        if gui:
            plt.ion()
        return super().__init__(*args, gui=gui, **kwargs)

    def plot_emission_yield(
        self,
        df: pd.DataFrame,
        *args,
        axes: Axes | None = None,
        population: ImplementedPop | None = None,
        **kwargs,
    ) -> Axes:
        """Plot :class:`.EmissionYield` data with |dfplot| method.

        Parameters
        ----------
        df :
            Dataframe holding data to plot.
        *args :
            Additional arguments passed to the |dfplot| method.
        axes :
            Axes to re-use if given.
        population :
            Type of population currently plotted. This is used to make the
            plot legends more precise.
        kwargs :
            Additional keyword arguments passed to the |dfplot| method.

        """
        if axes is not None:
            axes.set_prop_cycle(None)
        explicit = explicit_column_names(
            df.columns,
            population=population,
            emission_data_type="Emission Yield",
        )
        updated = df.rename(columns=explicit, inplace=False)
        axes = updated.plot(
            *args,
            x=explicit[col_energy],
            ax=axes,
            ylabel=md_ylabel["Emission Yield"],
            **kwargs,
        )
        assert isinstance(axes, Axes)
        return axes

    def plot_emission_energy_distribution(
        self,
        df: pd.DataFrame,
        *args,
        axes: Axes | None = None,
        population: ImplementedPop | None = None,
        **kwargs,
    ) -> Axes:
        """Plot :class:`.EmissionEnergyDistribution` data with |dfplot| method.

        Parameters
        ----------
        df :
            Dataframe holding data to plot.
        *args :
            Additional arguments passed to the |dfplot| method.
        axes :
            Axes to re-use if given.
        population :
            Type of population currently plotted. This is used to make the
            plot legends more precise.
        kwargs :
            Additional keyword arguments passed to the |dfplot| method.

        """
        if axes is not None:
            axes.set_prop_cycle(None)
        explicit = explicit_column_names(
            df.columns,
            population=population,
            emission_data_type="Emission Energy",
        )
        df.rename(columns=explicit, inplace=True)
        axes = df.plot(
            *args,
            x=explicit[col_energy],
            ax=axes,
            ylabel=md_ylabel["Emission Energy"],
            **kwargs,
        )
        assert isinstance(axes, Axes)
        return axes

    def plot_emission_angle_distribution(
        self,
        df: pd.DataFrame,
        *args,
        axes: Axes | None = None,
        population: ImplementedPop | None = None,
        **kwargs,
    ) -> Axes:
        """Plot the given emission angles distribution, return Axes object."""
        raise NotImplementedError(
            "Plotting emission angle distribution not implemented yet."
        )
